"""
    :codeauthor: Mike Wiebe <@mikewiebe>
"""

import pytest

import salt.modules.cp as cp_module
import salt.modules.file as file_module
import salt.modules.nxos as nxos_module
import salt.utils.nxos as nxos_utils
import salt.utils.pycrypto
from salt.exceptions import CommandExecutionError, SaltInvocationError
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, create_autospec, patch
from tests.support.unit import TestCase
from tests.unit.modules.nxos.nxos_config import (
    config_input_file,
    config_result,
    config_result_file,
    initial_config,
    initial_config_file,
    modified_config,
    modified_config_file,
    save_running_config,
    set_role,
    template_engine_file_str,
    template_engine_file_str_file,
    unset_role,
)
from tests.unit.modules.nxos.nxos_grains import n9k_grains
from tests.unit.modules.nxos.nxos_show_cmd_output import (
    n9k_get_user_output,
    n9k_show_user_account,
    n9k_show_user_account_list,
    n9k_show_ver,
    n9k_show_ver_int_list,
    n9k_show_ver_int_list_structured,
    n9k_show_ver_list,
)
from tests.unit.modules.nxos.nxos_show_run import (
    n9k_running_config,
    n9k_show_running_config_list,
    n9k_show_running_inc_username_list,
)


class NxosTestCase(TestCase, LoaderModuleMockMixin):
    """Test cases for salt.modules.nxos"""

    COPY_RS = "copy running-config startup-config"

    def setup_loader_modules(self):
        sendline = create_autospec(
            nxos_module.sendline, autospec=True, return_value={"command": "fake_output"}
        )
        return {nxos_module: {"__proxy__": {"nxos.sendline": sendline}}}

    @staticmethod
    def test_check_virtual():
        """UT: nxos module:check_virtual method - return value"""

        result = nxos_module.__virtual__()
        assert "nxos" in result

    def test_ping_proxy(self):
        """UT: nxos module:ping method - proxy"""
        with patch("salt.utils.platform.is_proxy", return_value=True, autospec=True):
            with patch.dict(
                nxos_module.__proxy__, {"nxos.ping": MagicMock(return_value=True)}
            ):
                result = nxos_module.ping()
                self.assertTrue(result)

    def test_ping_native_minion(self):
        """UT: nxos module:ping method - proxy"""

        with patch("salt.utils.platform.is_proxy", return_value=False, autospec=True):
            with patch.dict(
                nxos_module.__utils__, {"nxos.ping": MagicMock(return_value=True)}
            ):
                result = nxos_module.ping()
                self.assertTrue(result)

    def test_check_password_return_none(self):
        """UT: nxos module:check_password method - return None"""

        username = "admin"
        password = "foo"

        with patch("salt.modules.nxos.get_user", return_value=None, autospec=True):
            result = nxos_module.check_password(username, password, encrypted=False)
            self.assertIsNone(result)

    def test_check_password_password_nxos_comment(self):
        """UT: nxos module:check_password method - password_line has '!'"""

        username = "admin"
        password = "foo"

        with patch("salt.modules.nxos.get_user", return_value="!", autospec=True):
            result = nxos_module.check_password(username, password, encrypted=False)
            self.assertFalse(result)

    @pytest.mark.skipif(
        "sha256" not in salt.utils.pycrypto.methods,
        reason="compatible crypt method for fake data not available",
    )
    def test_check_password_password_encrypted_false(self):
        """UT: nxos module:check_password method - password is not encrypted"""

        username = "salt_test"
        password = "foobar123&"

        with patch(
            "salt.modules.nxos.get_user",
            return_value=n9k_get_user_output,
            autospec=True,
        ):
            result = nxos_module.check_password(username, password, encrypted=False)
            self.assertTrue(result)

    def test_check_password_password_encrypted_true(self):
        """UT: nxos module:check_password method - password is encrypted"""

        username = "salt_test"
        password = "$5$mkXh6O4T$YUVtA89HbXCnue63kgghPlaqPHyaXhdtxPBbPEHhbRC"

        with patch(
            "salt.modules.nxos.get_user",
            return_value=n9k_get_user_output,
            autospec=True,
        ):
            result = nxos_module.check_password(username, password, encrypted=True)
            self.assertTrue(result)

    def test_check_password_password_encrypted_true_negative(self):
        """UT: nxos module:check_password method - password is not encrypted"""

        username = "salt_test"
        password = "foobar123&"

        with patch(
            "salt.modules.nxos.get_user", return_value=n9k_running_config, autospec=True
        ):
            result = nxos_module.check_password(username, password, encrypted=True)
            self.assertFalse(result)

    def test_check_role_true(self):
        """UT: nxos module:check_role method - Role configured"""

        username = "salt_test"
        roles = ["network-admin", "dev-ops"]

        with patch("salt.modules.nxos.get_roles", return_value=roles, autospec=True):
            result = nxos_module.check_role(username, "dev-ops")
            self.assertTrue(result)

    def test_check_role_false(self):
        """UT: nxos module:check_role method - Role not configured"""

        username = "salt_test"
        roles = ["network-admin", "dev-ops"]

        with patch("salt.modules.nxos.get_roles", return_value=roles, autospec=True):
            result = nxos_module.check_role(username, "network-operator")
            self.assertFalse(result)

    def test_cmd_any_function(self):
        """UT: nxos module:cmd method - check_role function"""

        with patch.dict(
            nxos_module.__salt__,
            {
                "nxos.check_role": create_autospec(
                    nxos_module.check_role, return_value=True
                )
            },
        ):
            result = nxos_module.cmd(
                "check_role",
                "salt_test",
                "network-admin",
                encrypted=True,
                __pub_fun="nxos.cmd",
            )
            self.assertTrue(result)

    def test_cmd_function_absent(self):
        """UT: nxos module:cmd method - non existent function"""

        result = nxos_module.cmd(
            "cool_new_function", "salt_test", "network-admin", encrypted=True
        )
        self.assertFalse(result)

    def test_find_single_match(self):
        """UT: nxos module:test_find method - Find single match in running config"""

        find_pattern = "^vrf context testing$"
        find_string = "vrf context testing"

        with patch(
            "salt.modules.nxos.show_run", return_value=n9k_running_config, autospec=True
        ):
            result = nxos_module.find(find_pattern)
            self.assertIn(find_string, result)

    def test_find_multiple_matches(self):
        """UT: nxos module:test_find method - Find multiple matches in running config"""

        find_pattern = "^no logging.*$"
        find_string = "no logging event link-status enable"

        with patch(
            "salt.modules.nxos.show_run", return_value=n9k_running_config, autospec=True
        ):
            result = nxos_module.find(find_pattern)
            self.assertIn(find_string, result)
            self.assertEqual(len(result), 7)

    def test_get_roles_user_not_configured(self):
        """UT: nxos module:get_roles method - User not configured"""

        username = "salt_does_not_exist"
        user_info = ""

        with patch("salt.modules.nxos.get_user", return_value=user_info, autospec=True):
            result = nxos_module.get_roles(username)
            self.assertEqual(result, [])

    def test_get_roles_user_configured(self):
        """UT: nxos module:get_roles method - User configured"""

        username = "salt_test"
        expected_result = ["network-operator", "network-admin", "dev-ops"]

        with patch("salt.modules.nxos.get_user", return_value=username, autospec=True):
            for rv in [n9k_show_user_account, n9k_show_user_account_list]:
                with patch(
                    "salt.modules.nxos.sendline", return_value=rv, autospec=True
                ):
                    result = nxos_module.get_roles(username)
                    self.assertEqual(result.sort(), expected_result.sort())

    def test_get_roles_user_configured_no_role(self):
        """UT: nxos module:get_roles method - User configured no roles"""

        username = "salt_test"

        with patch("salt.modules.nxos.get_user", return_value=username, autospec=True):
            with patch("salt.modules.nxos.sendline", return_value="", autospec=True):
                result = nxos_module.get_roles(username)
                self.assertEqual(result, [])

    def test_get_user_configured(self):
        """UT: nxos module:get_user method - User configured"""

        username = "salt_test"
        expected_output = n9k_show_running_inc_username_list[0]

        for rv in [
            n9k_show_running_inc_username_list[0],
            n9k_show_running_inc_username_list,
        ]:
            with patch("salt.modules.nxos.sendline", return_value=rv, autospec=True):
                result = nxos_module.get_user(username)
                self.assertEqual(result, expected_output)

    def test_grains(self):
        """UT: nxos module:grains method"""

        nxos_module.DEVICE_DETAILS["grains_cache"] = {}
        expected_grains = {
            "software": {
                "BIOS": "version 08.36",
                "NXOS": "version 9.2(1)",
                "BIOS compile time": "06/07/2019",
                "NXOS image file is": "bootflash:///nxos.9.2.1.bin",
                "NXOS compile time": "7/17/2018 16:00:00 [07/18/2018 00:21:19]",
            },
            "hardware": {"Device name": "n9k-device", "bootflash": "53298520 kB"},
            "plugins": ["Core Plugin", "Ethernet Plugin"],
        }
        with patch.dict(
            nxos_module.__salt__,
            {
                "utils.nxos.system_info": create_autospec(
                    nxos_utils.system_info, return_value=n9k_grains
                )
            },
        ):
            with patch(
                "salt.modules.nxos.show_ver", return_value=n9k_show_ver, autospec=True
            ):
                result = nxos_module.grains()
                self.assertEqual(result, expected_grains)

    def test_grains_get_cache(self):
        """UT: nxos module:grains method"""

        expected_grains = {
            "software": {
                "BIOS": "version 08.36",
                "NXOS": "version 9.2(1)",
                "BIOS compile time": "06/07/2019",
                "NXOS image file is": "bootflash:///nxos.9.2.1.bin",
                "NXOS compile time": "7/17/2018 16:00:00 [07/18/2018 00:21:19]",
            },
            "hardware": {"Device name": "n9k-device", "bootflash": "53298520 kB"},
            "plugins": ["Core Plugin", "Ethernet Plugin"],
        }
        nxos_module.DEVICE_DETAILS["grains_cache"] = expected_grains
        with patch.dict(
            nxos_module.__salt__,
            {
                "utils.nxos.system_info": create_autospec(
                    nxos_utils.system_info, return_value=n9k_grains
                )
            },
        ):
            with patch(
                "salt.modules.nxos.show_ver", return_value=n9k_show_ver, autospec=True
            ):
                result = nxos_module.grains()
                self.assertEqual(result, expected_grains)

    def test_grains_refresh(self):
        """UT: nxos module:grains_refresh method"""

        expected_grains = {
            "software": {
                "BIOS": "version 08.36",
                "NXOS": "version 9.2(1)",
                "BIOS compile time": "06/07/2019",
                "NXOS image file is": "bootflash:///nxos.9.2.1.bin",
                "NXOS compile time": "7/17/2018 16:00:00 [07/18/2018 00:21:19]",
            },
            "hardware": {"Device name": "n9k-device", "bootflash": "53298520 kB"},
            "plugins": ["Core Plugin", "Ethernet Plugin"],
        }

        with patch(
            "salt.modules.nxos.grains", return_value=expected_grains, autospec=True
        ):
            result = nxos_module.grains_refresh()
            self.assertEqual(result, expected_grains)

    def test_system_info(self):
        """UT: nxos module:system_info method"""

        expected_grains = {
            "software": {
                "BIOS": "version 08.36",
                "NXOS": "version 9.2(1)",
                "BIOS compile time": "06/07/2019",
                "NXOS image file is": "bootflash:///nxos.9.2.1.bin",
                "NXOS compile time": "7/17/2018 16:00:00 [07/18/2018 00:21:19]",
            },
            "hardware": {"Device name": "n9k-device", "bootflash": "53298520 kB"},
            "plugins": ["Core Plugin", "Ethernet Plugin"],
        }
        with patch.dict(
            nxos_module.__salt__,
            {
                "utils.nxos.system_info": create_autospec(
                    nxos_utils.system_info, return_value=n9k_grains
                )
            },
        ):
            with patch(
                "salt.modules.nxos.show_ver", return_value=n9k_show_ver, autospec=True
            ):
                result = nxos_module.system_info()
                self.assertEqual(result, expected_grains)

    def test_sendline_invalid_method(self):
        """UT: nxos module:sendline method - invalid method"""

        command = "show version"
        method = "invalid"

        # Execute the function under test
        result = nxos_module.sendline(command, method)

        self.assertIn("INPUT ERROR", result)

    def test_sendline_valid_method_proxy(self):
        """UT: nxos module:sendline method - valid method over proxy"""

        command = "show version"
        method = "cli_show_ascii"

        with patch("salt.utils.platform.is_proxy", return_value=True, autospec=True):
            with patch.dict(
                nxos_module.__proxy__,
                {"nxos.sendline": MagicMock(return_value=n9k_show_ver)},
            ):
                result = nxos_module.sendline(command, method)
                self.assertIn(n9k_show_ver, result)

    def test_sendline_valid_method_nxapi_uds(self):
        """UT: nxos module:sendline method - valid method over nxapi uds"""

        command = "show version"
        method = "cli_show_ascii"

        with patch("salt.utils.platform.is_proxy", MagicMock(return_value=False)):
            with patch(
                "salt.modules.nxos._nxapi_request",
                return_value=n9k_show_ver,
                autospec=True,
            ):
                result = nxos_module.sendline(command, method)
                self.assertIn(n9k_show_ver, result)

    def test_show_raw_text_invalid(self):
        """UT: nxos module:show method - invalid argument"""

        command = "show version"
        raw_text = "invalid"

        result = nxos_module.show(command, raw_text)
        self.assertIn("INPUT ERROR", result)

    def test_show_raw_text_true(self):
        """UT: nxos module:show method - raw_test true"""

        command = "show version"
        raw_text = True

        with patch(
            "salt.modules.nxos.sendline", autospec=True, return_value=n9k_show_ver
        ):
            result = nxos_module.show(command, raw_text)
            self.assertEqual(result, n9k_show_ver)

    def test_show_raw_text_true_multiple_commands(self):
        """UT: nxos module:show method - raw_test true multiple commands"""

        command = "show bgp sessions ; show processes"
        raw_text = True
        data = ["bgp_session_data", "process_data"]

        with patch("salt.modules.nxos.sendline", autospec=True, return_value=data):
            result = nxos_module.show(command, raw_text)
            self.assertEqual(result, data)

    def test_show_nxapi(self):
        """UT: nxos module:show method - nxapi returns info as list"""

        command = "show version; show interface eth1/1"
        raw_text = True

        with patch(
            "salt.modules.nxos.sendline",
            autospec=True,
            return_value=n9k_show_ver_int_list,
        ):
            result = nxos_module.show(command, raw_text)
            self.assertEqual(result[0], n9k_show_ver_int_list[0])
            self.assertEqual(result[1], n9k_show_ver_int_list[1])

    def test_show_nxapi_structured(self):
        """UT: nxos module:show method - nxapi returns info as list"""

        command = "show version; show interface eth1/1"
        raw_text = False

        with patch(
            "salt.modules.nxos.sendline",
            autospec=True,
            return_value=n9k_show_ver_int_list_structured,
        ):
            result = nxos_module.show(command, raw_text)
            self.assertEqual(result[0], n9k_show_ver_int_list_structured[0])
            self.assertEqual(result[1], n9k_show_ver_int_list_structured[1])

    def test_show_run(self):
        """UT: nxos module:show_run method"""

        expected_output = n9k_show_running_config_list[0]

        for rv in [n9k_show_running_config_list[0], n9k_show_running_config_list]:
            with patch("salt.modules.nxos.sendline", autospec=True, return_value=rv):
                result = nxos_module.show_run()
                self.assertEqual(result, expected_output)

    def test_show_ver(self):
        """UT: nxos module:show_ver method"""

        expected_output = n9k_show_ver_list[0]

        for rv in [n9k_show_ver_list[0], n9k_show_ver_list]:
            with patch("salt.modules.nxos.sendline", autospec=True, return_value=rv):
                result = nxos_module.show_ver()
                self.assertEqual(result, expected_output)

    def test_add_config(self):
        """UT: nxos module:add_config method"""

        expected_output = "COMMAND_LIST: feature bgp"

        with patch(
            "salt.modules.nxos.config", autospec=True, return_value=expected_output
        ):
            result = nxos_module.add_config("feature bgp")
            self.assertEqual(result, expected_output)

    def test_config_commands(self):
        """UT: nxos module:config method - Using commands arg"""

        commands = ["no feature ospf", ["no feature ospf"]]
        expected_output = (
            "COMMAND_LIST: no feature ospf\n\n--- \n+++ \n@@ -19,7 +19,6 @@\n feature"
            " bash-shell\n cfs eth distribute\n feature ngmvpn\n-feature ospf\n feature"
            " pim\n feature lldp\n \n"
        )

        for cmd_set in commands:
            with patch(
                "salt.modules.nxos.sendline",
                autospec=True,
                side_effect=[initial_config, modified_config],
            ):
                mock_cmd = create_autospec(
                    file_module.apply_template_on_contents,
                    return_value=template_engine_file_str,
                )
                with patch.dict(
                    nxos_module.__salt__, {"file.apply_template_on_contents": mock_cmd}
                ):
                    with patch(
                        "salt.modules.nxos._configure_device",
                        autospec=True,
                        return_value=config_result,
                    ):
                        result = nxos_module.config(cmd_set, save_config=False)
                        self.assertEqual(result, expected_output)

    def test_config_commands_template_none(self):
        """UT: nxos module:config method - Template engine is None"""

        commands = ["no feature ospf", ["no feature ospf"]]
        expected_output = (
            "COMMAND_LIST: no feature ospf\n\n--- \n+++ \n@@ -19,7 +19,6 @@\n feature"
            " bash-shell\n cfs eth distribute\n feature ngmvpn\n-feature ospf\n feature"
            " pim\n feature lldp\n \n"
        )

        for cmd_set in commands:
            with patch(
                "salt.modules.nxos.sendline",
                autospec=True,
                side_effect=[initial_config, modified_config],
            ):
                mock_cmd = create_autospec(
                    file_module.apply_template_on_contents,
                    return_value=template_engine_file_str,
                )
                with patch.dict(
                    nxos_module.__salt__, {"file.apply_template_on_contents": mock_cmd}
                ):
                    with patch(
                        "salt.modules.nxos._configure_device",
                        autospec=True,
                        return_value=config_result,
                    ):
                        result = nxos_module.config(cmd_set, template_engine=None)
                        self.assertEqual(result, expected_output)

    def test_config_commands_string(self):
        """UT: nxos module:config method - Using commands arg and output is string"""

        commands = "no feature ospf"
        expected_output = (
            "COMMAND_LIST: no feature ospf\n\n--- \n+++ \n@@ -19,7 +19,6 @@\n feature"
            " bash-shell\n cfs eth distribute\n feature ngmvpn\n-feature ospf\n feature"
            " pim\n feature lldp\n \n"
        )

        with patch(
            "salt.modules.nxos.sendline",
            autospec=True,
            side_effect=[initial_config[0], modified_config[0]],
        ):
            mock_cmd = create_autospec(
                file_module.apply_template_on_contents,
                return_value=template_engine_file_str,
            )
            with patch.dict(
                nxos_module.__salt__, {"file.apply_template_on_contents": mock_cmd}
            ):
                with patch(
                    "salt.modules.nxos._configure_device",
                    autospec=True,
                    return_value=config_result,
                ):
                    result = nxos_module.config(commands)
                    self.assertEqual(result, expected_output)

    def test_config_file(self):
        """UT: nxos module:config method - Using config_file arg"""

        config_file = "salt://bgp_config.txt"
        expected_output = (
            "COMMAND_LIST: feature bgp ; ! ; router bgp 55 ; address-family ipv4"
            " unicast ; no client-to-client reflection ; additional-paths send\n\n---"
            " \n+++ \n@@ -19,6 +19,7 @@\n feature bash-shell\n cfs eth distribute\n"
            " feature ngmvpn\n+feature bgp\n feature pim\n feature lldp\n \n@@ -233,6"
            " +234,10 @@\n line console\n line vty\n boot nxos"
            " bootflash:/nxos.9.2.4.bin \n+router bgp 55\n+  address-family ipv4"
            " unicast\n+    no client-to-client reflection\n+    additional-paths"
            " send\n \n no logging logfile\n no logging monitor\n"
        )

        with patch(
            "salt.modules.nxos.sendline",
            autospec=True,
            side_effect=[initial_config_file, modified_config_file],
        ):
            mock_cmd = create_autospec(
                cp_module.get_file_str, return_value=config_input_file
            )
            with patch.dict(nxos_module.__salt__, {"cp.get_file_str": mock_cmd}):
                mock_cmd = create_autospec(
                    file_module.apply_template_on_contents,
                    return_value=template_engine_file_str_file,
                )
                with patch.dict(
                    nxos_module.__salt__, {"file.apply_template_on_contents": mock_cmd}
                ):
                    with patch(
                        "salt.modules.nxos._configure_device",
                        autospec=True,
                        return_value=config_result_file,
                    ):
                        result = nxos_module.config(config_file=config_file)
                        self.assertEqual(result, expected_output)

    def test_config_file_error1(self):
        """UT: nxos module:config method - Error file not found"""

        config_file = "salt://bgp_config.txt"

        with patch(
            "salt.modules.nxos.sendline",
            autospec=True,
            side_effect=[initial_config_file, modified_config_file],
        ):
            mock_cmd = create_autospec(cp_module.get_file_str, return_value=False)
            with patch.dict(nxos_module.__salt__, {"cp.get_file_str": mock_cmd}):
                mock_cmd = create_autospec(
                    file_module.apply_template_on_contents,
                    return_value=template_engine_file_str_file,
                )
                with patch.dict(
                    nxos_module.__salt__, {"file.apply_template_on_contents": mock_cmd}
                ):
                    with patch(
                        "salt.modules.nxos._configure_device",
                        autospec=True,
                        return_value=config_result_file,
                    ):
                        with self.assertRaises(CommandExecutionError):
                            nxos_module.config(config_file=config_file)

    def test_config_nxos_error_ssh(self):
        """UT: nxos module:config method - nxos device error over ssh transport"""

        commands = ["feature bgp", "router bgp 57"]
        config_result = [
            ["feature bgp", "router bgp 57"],
            "bgp instance is already running; Tag is 55",
        ]
        expected_output = (
            "COMMAND_LIST: feature bgp ; router bgp 57\nbgp instance is already"
            " running; Tag is 55\n--- \n+++ \n@@ -19,7 +19,6 @@\n feature bash-shell\n"
            " cfs eth distribute\n feature ngmvpn\n-feature ospf\n feature pim\n"
            " feature lldp\n \n"
        )

        with patch(
            "salt.modules.nxos.sendline",
            autospec=True,
            side_effect=[initial_config[0], modified_config[0]],
        ):
            mock_cmd = create_autospec(
                file_module.apply_template_on_contents,
                return_value=template_engine_file_str,
            )
            with patch.dict(
                nxos_module.__salt__, {"file.apply_template_on_contents": mock_cmd}
            ):
                with patch(
                    "salt.modules.nxos._configure_device",
                    autospec=True,
                    return_value=config_result,
                ):
                    result = nxos_module.config(commands)
                    self.assertEqual(result, expected_output)

    def test_commands_error(self):
        """UT: nxos module:config method - Mandatory arg commands not specified"""

        commands = None

        with patch(
            "salt.modules.nxos.sendline",
            autospec=True,
            side_effect=[initial_config_file, modified_config_file],
        ):
            mock_cmd = create_autospec(cp_module.get_file_str, return_value=False)
            with patch.dict(nxos_module.__salt__, {"cp.get_file_str": mock_cmd}):
                mock_cmd = create_autospec(
                    file_module.apply_template_on_contents,
                    return_value=template_engine_file_str_file,
                )
                with patch.dict(
                    nxos_module.__salt__, {"file.apply_template_on_contents": mock_cmd}
                ):
                    with patch(
                        "salt.modules.nxos._configure_device",
                        autospec=True,
                        return_value=config_result_file,
                    ):
                        with self.assertRaises(CommandExecutionError):
                            nxos_module.config(commands=commands)

    def test_config_file_error2(self):
        """UT: nxos module:config method - Mandatory arg config_file not specified"""

        config_file = None

        with patch(
            "salt.modules.nxos.sendline",
            autospec=True,
            side_effect=[initial_config_file, modified_config_file],
        ):
            mock_cmd = create_autospec(cp_module.get_file_str, return_value=False)
            with patch.dict(nxos_module.__salt__, {"cp.get_file_str": mock_cmd}):
                mock_cmd = create_autospec(
                    file_module.apply_template_on_contents,
                    return_value=template_engine_file_str_file,
                )
                with patch.dict(
                    nxos_module.__salt__, {"file.apply_template_on_contents": mock_cmd}
                ):
                    with patch(
                        "salt.modules.nxos._configure_device",
                        autospec=True,
                        return_value=config_result_file,
                    ):
                        with self.assertRaises(CommandExecutionError):
                            nxos_module.config(config_file=config_file)

    def test_delete_config(self):
        """UT: nxos module:delete_config method"""

        for lines in ["feature bgp", ["feature bgp"]]:
            with patch("salt.modules.nxos.config", autospec=True):
                result = nxos_module.delete_config(lines)
                nxos_module.config.assert_called_with(["no feature bgp"])
                self.assertEqual(result, nxos_module.config.return_value)

    def test_remove_user(self):
        """UT: nxos module:remove_user method"""

        with patch("salt.modules.nxos.config", autospec=True):
            result = nxos_module.remove_user("salt_test")
            nxos_module.config.assert_called_with("no username salt_test")
            self.assertEqual(result, nxos_module.config.return_value)

    def test_replace(self):
        """UT: nxos module:replace method"""

        old_value = "feature bgp"
        new_value = "feature ospf"

        with patch(
            "salt.modules.nxos.show_run",
            autospec=True,
            return_value=n9k_show_running_config_list[0],
        ):
            with patch(
                "salt.modules.nxos.delete_config", autospec=True, return_value=None
            ):
                with patch(
                    "salt.modules.nxos.config", autospec=True, return_value=None
                ):
                    result = nxos_module.replace(old_value, new_value)
                    self.assertEqual(result["old"], ["feature bgp"])
                    self.assertEqual(result["new"], ["feature ospf"])

    def test_replace_full_match_true(self):
        """UT: nxos module:replace method - full match true"""

        old_value = "feature bgp"
        new_value = "feature ospf"

        with patch(
            "salt.modules.nxos.show_run",
            autospec=True,
            return_value=n9k_show_running_config_list[0],
        ):
            with patch(
                "salt.modules.nxos.delete_config", autospec=True, return_value=None
            ):
                with patch(
                    "salt.modules.nxos.config", autospec=True, return_value=None
                ):
                    result = nxos_module.replace(old_value, new_value, full_match=True)
                    self.assertEqual(result["old"], ["feature bgp"])
                    self.assertEqual(result["new"], ["feature ospf"])

    def test_replace_no_match(self):
        """UT: nxos module:replace method - no match"""

        old_value = "feature does_not_exist"
        new_value = "feature ospf"

        with patch(
            "salt.modules.nxos.show_run",
            autospec=True,
            return_value=n9k_show_running_config_list[0],
        ):
            with patch(
                "salt.modules.nxos.delete_config", autospec=True, return_value=None
            ):
                with patch(
                    "salt.modules.nxos.config", autospec=True, return_value=None
                ):
                    result = nxos_module.replace(old_value, new_value)
                    self.assertEqual(result["old"], [])
                    self.assertEqual(result["new"], [])

    def test_save_running_config(self):
        """UT: nxos module:save_running_config method"""

        with patch(
            "salt.modules.nxos.config", autospec=True, return_value=save_running_config
        ):
            result = nxos_module.save_running_config()
            self.assertEqual(result, save_running_config)

    def test_set_password_enc_false_cs_none(self):
        """UT: nxos module:set_password method - encrypted False, crypt_salt None"""

        username = "devops"
        password = "test123TMM^&"
        hashed_pass = "$5$ZcZqm15X$exHN2m6yrPKpYhGArK3Vml3ZjNbJaJYdzWyf0fp1Up2"
        config_line = (
            "username devops password 5"
            " $5$ZcZqm15X$exHN2m6yrPKpYhGArK3Vml3ZjNbJaJYdzWyf0fp1Up2"
        )

        with patch("salt.modules.nxos.get_user", autospec=True):
            with patch(
                "salt.modules.nxos.gen_hash", autospec=True, return_value=hashed_pass
            ):
                with patch(
                    "salt.modules.nxos.config",
                    autospec=True,
                    return_value="password_set",
                ) as config:
                    result = nxos_module.set_password(username, password)
                    config.assert_called_with(config_line)
                    self.assertEqual("password_set", result)

    def test_set_password_enc_false_cs_set(self):
        """UT: nxos module:set_password method - encrypted False, crypt_salt set"""

        username = "devops"
        password = "test123TMM^&"
        crypt_salt = "ZcZqm15X"
        hashed_pass = "$5$ZcZqm15X$exHN2m6yrPKpYhGArK3Vml3ZjNbJaJYdzWyf0fp1Up2"
        config_line = (
            "username devops password 5"
            " $5$ZcZqm15X$exHN2m6yrPKpYhGArK3Vml3ZjNbJaJYdzWyf0fp1Up2"
        )

        with patch("salt.modules.nxos.get_user", autospec=True):
            with patch(
                "salt.modules.nxos.gen_hash", autospec=True, return_value=hashed_pass
            ):
                with patch(
                    "salt.modules.nxos.config",
                    autospec=True,
                    return_value="password_set",
                ) as config:
                    result = nxos_module.set_password(
                        username, password, crypt_salt=crypt_salt
                    )
                    config.assert_called_with(config_line)
                    self.assertEqual("password_set", result)

    def test_set_password_enc_true(self):
        """UT: nxos module:set_password method - encrypted True"""

        username = "devops"
        password = "test123TMM^&"
        hashed_pass = "$5$ZcZqm15X$exHN2m6yrPKpYhGArK3Vml3ZjNbJaJYdzWyf0fp1Up2"
        config_line = "username devops password 5 test123TMM^&"

        with patch("salt.modules.nxos.get_user", autospec=True):
            with patch(
                "salt.modules.nxos.gen_hash", autospec=True, return_value=hashed_pass
            ):
                with patch(
                    "salt.modules.nxos.config",
                    autospec=True,
                    return_value="password_set",
                ) as config:
                    result = nxos_module.set_password(
                        username, password, encrypted=True
                    )
                    config.assert_called_with(config_line)
                    self.assertEqual("password_set", result)

    def test_set_password_role_none(self):
        """UT: nxos module:set_password method - role none"""

        username = "devops"
        password = "test123TMM^&"
        hashed_pass = "$5$ZcZqm15X$exHN2m6yrPKpYhGArK3Vml3ZjNbJaJYdzWyf0fp1Up2"
        config_line = "username devops password 5 test123TMM^& role devops"

        with patch("salt.modules.nxos.get_user", autospec=True):
            with patch(
                "salt.modules.nxos.gen_hash", autospec=True, return_value=hashed_pass
            ):
                with patch(
                    "salt.modules.nxos.config",
                    autospec=True,
                    return_value="password_set",
                ) as config:
                    # Execute the function under test
                    result = nxos_module.set_password(
                        username, password, encrypted=True, role="devops"
                    )
                    config.assert_called_with(config_line)
                    self.assertEqual("password_set", result)

    def test_set_password_blowfish_crypt(self):
        """UT: nxos module:set_password method - role none"""

        with self.assertRaises(SaltInvocationError):
            nxos_module.set_password(
                "username", "password", encrypted=True, algorithm="blowfish"
            )

    def test_set_role(self):
        """UT: nxos module:save_running_config method"""

        username = "salt_test"
        role = "vdc-admin"

        with patch("salt.modules.nxos.config", autospec=True, return_value=set_role):
            result = nxos_module.set_role(username, role)
            self.assertEqual(result, set_role)

    def test_unset_role(self):
        """UT: nxos module:save_running_config method"""

        username = "salt_test"
        role = "vdc-admin"

        with patch("salt.modules.nxos.config", autospec=True, return_value=unset_role):
            result = nxos_module.unset_role(username, role)
            self.assertEqual(result, unset_role)

    def test_configure_device(self):
        """UT: nxos module:_configure_device method"""

        with patch("salt.utils.platform.is_proxy", autospec=True, return_value=True):
            with patch.dict(
                nxos_module.__proxy__,
                {"nxos.proxy_config": MagicMock(return_value="configured")},
            ):
                result = nxos_module._configure_device("feature bgp")
                self.assertEqual(result, "configured")

        with patch("salt.utils.platform.is_proxy", autospec=True, return_value=False):
            with patch.object(
                nxos_module, "_nxapi_config", MagicMock(return_value="configured")
            ):
                nxos_module._configure_device("feature bgp")
                self.assertEqual(result, "configured")

    def test_nxapi_config(self):
        """UT: nxos module:_nxapi_config method"""

        mock_cmd = MagicMock(return_value={"nxos": {"save_config": False}})
        with patch.dict(nxos_module.__salt__, {"config.get": mock_cmd}):
            with patch(
                "salt.modules.nxos._nxapi_request",
                return_value="router_data",
                autospec=True,
            ):
                result = nxos_module._nxapi_config("show version")
                self.assertEqual(result, [["show version"], "router_data"])

    def test_nxapi_config_failure(self):
        """UT: nxos module:_nxapi_config method"""

        side_effect = ["Failure", "saved_data"]

        mock_cmd = MagicMock(return_value={"nxos": {"save_config": True}})
        with patch.dict(nxos_module.__salt__, {"config.get": mock_cmd}):
            with patch(
                "salt.modules.nxos._nxapi_request",
                side_effect=side_effect,
                autospec=True,
            ):
                result = nxos_module._nxapi_config("show bad_command")
                self.assertEqual(result, [["show bad_command"], "Failure"])

    def test_nxapi_request_proxy(self):
        """UT: nxos module:_nxapi_request method - proxy"""

        with patch("salt.utils.platform.is_proxy", autospec=True, return_value=True):
            mock_request = create_autospec(
                nxos_utils.nxapi_request, return_value="router_data"
            )
            with patch.dict(
                nxos_module.__proxy__, {"nxos._nxapi_request": mock_request}
            ):
                result = nxos_module._nxapi_request("show version")
                self.assertEqual(result, "router_data")

    def test_nxapi_request_no_proxy(self):
        """UT: nxos module:_nxapi_request method - no proxy"""

        with patch("salt.utils.platform.is_proxy", autospec=True, return_value=False):
            mock_cmd = MagicMock(return_value={"nxos": {"save_config": False}})
            with patch.dict(nxos_module.__salt__, {"config.get": mock_cmd}):
                mock_request = create_autospec(nxos_utils.nxapi_request)
                with patch.dict(
                    nxos_module.__utils__, {"nxos.nxapi_request": mock_request}
                ):
                    result = nxos_module._nxapi_request("show version")
                    self.assertEqual(result, mock_request.return_value)
                    mock_request.assert_called_with(
                        "show version", "cli_conf", **mock_cmd.return_value
                    )



import shutil
import tempfile
from pathlib import Path

import pytest

import salt.pillar.hg_pillar as hg_pillar

try:
    import hglib

    HAS_HG = True
except ImportError:
    HAS_HG = False


@pytest.fixture(scope="module")
def configure_loader_modules(master_opts):
    yield {hg_pillar: {"__opts__": master_opts}}


@pytest.fixture
def hg_setup_and_teardown():
    """
    build up and tear down hg repos to test with.
    """
    sourcedirPath = Path(__file__).resolve().parent.joinpath("files")
    tempdir = tempfile.TemporaryDirectory()
    tempsubdir = tempdir.name / Path("test2/")
    tempsubdir2 = tempdir.name / Path("subdir/")
    tempsubdir3 = tempdir.name / Path("subdir/test2/")
    tempsubdir.mkdir()
    tempsubdir2.mkdir()
    tempsubdir3.mkdir()
    tempdirPath = Path(tempdir.name)
    filessrc = [
        Path("top.sls"),
        Path("test.sls"),
        Path("test2/init.sls"),
    ]
    for fnd in filessrc:
        to = tempdirPath / fnd
        to2 = tempsubdir2 / fnd
        frm = sourcedirPath / fnd
        shutil.copy(frm.as_posix(), to.as_posix())
        shutil.copy(frm.as_posix(), to2.as_posix())
    hglib.init(bytes(tempdirPath.as_posix(), encoding="utf8"))
    repo = hglib.open(bytes(tempdirPath.as_posix(), encoding="utf8"))
    repo.add(bytes(tempdirPath.as_posix(), encoding="utf8"))
    repo.commit(b"init commit", user="test")
    repo.tag(b"test", user="test")
    repo.branch(b"test")
    repo.commit(b"create test branch", user="test")
    repo.bookmark(b"bookmark_test")
    yield tempdirPath.as_uri()
    tempdir.cleanup()


@pytest.mark.slow_test
@pytest.mark.skip_on_windows(
    reason="just testing if this or hgfs causes the issue with total crash"
)
def test_ext_pillar(hg_setup_and_teardown):
    data = hg_pillar.ext_pillar("*", None, hg_setup_and_teardown)
    assert data == {"testinfo": "info", "testinfo2": "info"}
    data = hg_pillar.ext_pillar("test", None, hg_setup_and_teardown)
    assert data == {"testinfo": "info", "testinfo2": "info"}
    data = hg_pillar.ext_pillar("*", None, hg_setup_and_teardown, root="subdir")
    assert data == {"testinfo": "info", "testinfo2": "info"}



"""
SaltStack Extend
~~~~~~~~~~~~~~~~

A templating tool for extending SaltStack.

Takes a template directory and merges it into a SaltStack source code
directory. This tool uses Jinja2 for templating.

This tool is accessed using `salt-extend`

    :codeauthor: Anthony Shaw <anthonyshaw@apache.org>
"""

import logging
import os
import shutil
import sys
import tempfile
from collections import OrderedDict
from datetime import date

from jinja2 import Template

import salt.utils.files
import salt.version
from salt.serializers.yaml import deserialize

log = logging.getLogger(__name__)

try:
    import click

    HAS_CLICK = True
except ImportError as ie:
    HAS_CLICK = False

TEMPLATE_FILE_NAME = "template.yml"


def _get_template(path, option_key):
    """
    Get the contents of a template file and provide it as a module type

    :param path: path to the template.yml file
    :type  path: ``str``

    :param option_key: The unique key of this template
    :type  option_key: ``str``

    :returns: Details about the template
    :rtype: ``tuple``
    """
    with salt.utils.files.fopen(path, "r") as template_f:
        template = deserialize(template_f)
        info = (option_key, template.get("description", ""), template)
    return info


def _fetch_templates(src):
    """
    Fetch all of the templates in the src directory

    :param src: The source path
    :type  src: ``str``

    :rtype: ``list`` of ``tuple``
    :returns: ``list`` of ('key', 'description')
    """
    templates = []
    log.debug("Listing contents of %s", src)
    for item in os.listdir(src):
        s = os.path.join(src, item)
        if os.path.isdir(s):
            template_path = os.path.join(s, TEMPLATE_FILE_NAME)
            if os.path.isfile(template_path):
                templates.append(_get_template(template_path, item))
            else:
                log.debug(
                    "Directory does not contain %s %s",
                    template_path,
                    TEMPLATE_FILE_NAME,
                )
    return templates


def _mergetree(src, dst):
    """
    Akin to shutils.copytree but over existing directories, does a recursive merge copy.

    :param src: The source path
    :type  src: ``str``

    :param dst: The destination path
    :type  dst: ``str``
    """
    for item in os.listdir(src):
        s = os.path.join(src, item)
        d = os.path.join(dst, item)
        if os.path.isdir(s):
            log.info("Copying folder %s to %s", s, d)
            if os.path.exists(d):
                _mergetree(s, d)
            else:
                shutil.copytree(s, d)
        else:
            log.info("Copying file %s to %s", s, d)
            shutil.copy2(s, d)


def _mergetreejinja(src, dst, context):
    """
    Merge directory A to directory B, apply Jinja2 templating to both
    the file/folder names AND to the contents of the files

    :param src: The source path
    :type  src: ``str``

    :param dst: The destination path
    :type  dst: ``str``

    :param context: The dictionary to inject into the Jinja template as context
    :type  context: ``dict``
    """
    for item in os.listdir(src):
        s = os.path.join(src, item)
        d = os.path.join(dst, item)
        if os.path.isdir(s):
            log.info("Copying folder %s to %s", s, d)
            if os.path.exists(d):
                _mergetreejinja(s, d, context)
            else:
                os.mkdir(d)
                _mergetreejinja(s, d, context)
        else:
            if item != TEMPLATE_FILE_NAME:
                d = Template(d).render(context)
                log.info("Copying file %s to %s", s, d)
                with salt.utils.files.fopen(s, "r") as source_file:
                    src_contents = salt.utils.stringutils.to_unicode(source_file.read())
                    dest_contents = Template(src_contents).render(context)
                with salt.utils.files.fopen(d, "w") as dest_file:
                    dest_file.write(salt.utils.stringutils.to_str(dest_contents))


def _prompt_user_variable(var_name, default_value):
    """
    Prompt the user to enter the value of a variable

    :param var_name: The question to ask the user
    :type  var_name: ``str``

    :param default_value: The default value
    :type  default_value: ``str``

    :rtype: ``str``
    :returns: the value from the user
    """
    return click.prompt(var_name, default=default_value)


def _prompt_choice(var_name, options):
    """
    Prompt the user to choose between a list of options, index each one by adding an enumerator
    based on https://github.com/audreyr/cookiecutter/blob/master/cookiecutter/prompt.py#L51

    :param var_name: The question to ask the user
    :type  var_name: ``str``

    :param options: A list of options
    :type  options: ``list`` of ``tupple``

    :rtype: ``tuple``
    :returns: The selected user
    """
    choice_map = OrderedDict(
        (f"{i}", value) for i, value in enumerate(options, 1) if value[0] != "test"
    )
    choices = choice_map.keys()
    default = "1"

    choice_lines = [f"{c[0]} - {c[1][0]} - {c[1][1]}" for c in choice_map.items()]
    prompt = "\n".join(
        (
            f"Select {var_name}:",
            "\n".join(choice_lines),
            "Choose from {}".format(", ".join(choices)),
        )
    )

    user_choice = click.prompt(prompt, type=click.Choice(choices), default=default)
    return choice_map[user_choice]


def apply_template(template_dir, output_dir, context):
    """
    Apply the template from the template directory to the output
    using the supplied context dict.

    :param src: The source path
    :type  src: ``str``

    :param dst: The destination path
    :type  dst: ``str``

    :param context: The dictionary to inject into the Jinja template as context
    :type  context: ``dict``
    """
    _mergetreejinja(template_dir, output_dir, context)


def run(
    extension=None,
    name=None,
    description=None,
    salt_dir=None,
    merge=False,
    temp_dir=None,
):
    """
    A template factory for extending the salt ecosystem

    :param extension: The extension type, e.g. 'module', 'state', if omitted, user will be prompted
    :type  extension: ``str``

    :param name: Python-friendly name for the module, if omitted, user will be prompted
    :type  name: ``str``

    :param description: A description of the extension, if omitted, user will be prompted
    :type  description: ``str``

    :param salt_dir: The targeted Salt source directory
    :type  salt_dir: ``str``

    :param merge: Merge with salt directory, `False` to keep separate, `True` to merge trees.
    :type  merge: ``bool``

    :param temp_dir: The directory for generated code, if omitted, system temp will be used
    :type  temp_dir: ``str``
    """
    if not HAS_CLICK:
        print("click is not installed, please install using pip")
        sys.exit(1)

    if salt_dir is None:
        salt_dir = "."

    MODULE_OPTIONS = _fetch_templates(os.path.join(salt_dir, "templates"))

    if extension is None:
        print("Choose which type of extension you are developing for SaltStack")
        extension_type = "Extension type"
        chosen_extension = _prompt_choice(extension_type, MODULE_OPTIONS)
    else:
        if extension not in list(zip(*MODULE_OPTIONS))[0]:
            print("Module extension option not valid")
            sys.exit(1)

        chosen_extension = [m for m in MODULE_OPTIONS if m[0] == extension][0]

    extension_type = chosen_extension[0]
    extension_context = chosen_extension[2]

    if name is None:
        print("Enter the short name for the module (e.g. mymodule)")
        name = _prompt_user_variable("Module name", "")

    if description is None:
        description = _prompt_user_variable("Short description of the module", "")

    template_dir = f"templates/{extension_type}"
    module_name = name

    param_dict = {
        "version": salt.version.SaltStackVersion.next_release().name,
        "module_name": module_name,
        "short_description": description,
        "release_date": date.today().strftime("%Y-%m-%d"),
        "year": date.today().strftime("%Y"),
    }

    # get additional questions from template
    additional_context = {}
    for key, val in extension_context.get("questions", {}).items():
        # allow templates to be used in default values.
        default = Template(val.get("default", "")).render(param_dict)

        prompt_var = _prompt_user_variable(val["question"], default)
        additional_context[key] = prompt_var

    context = param_dict.copy()
    context.update(extension_context)
    context.update(additional_context)

    if temp_dir is None:
        temp_dir = tempfile.mkdtemp()

    apply_template(template_dir, temp_dir, context)

    if not merge:
        path = temp_dir
    else:
        _mergetree(temp_dir, salt_dir)
        path = salt_dir

    log.info("New module stored in %s", path)
    return path


if __name__ == "__main__":
    run()



import salt.pillar.pepa as pepa

try:
    from collections import OrderedDict
except ImportError:
    from collections import OrderedDict


def test_repeated_keys():
    expected_result = {
        "foo": {
            "bar": {
                "foo": True,
                "baz": True,
            },
        },
    }
    data = OrderedDict(
        [
            ("foo..bar..foo", True),
            ("foo..bar..baz", True),
        ]
    )
    result = pepa.key_value_to_tree(data)
    assert result == expected_result



import getpass
import logging
import os

import salt.utils.files
import salt.utils.stringutils
from salt.exceptions import InvalidKeyError

try:
    from M2Crypto import BIO, EVP, RSA

    HAS_M2 = True
except ImportError:
    HAS_M2 = False

if not HAS_M2:
    try:
        from Cryptodome import Random
        from Cryptodome.Cipher import AES, PKCS1_OAEP
        from Cryptodome.Cipher import PKCS1_v1_5 as PKCS1_v1_5_CIPHER
        from Cryptodome.Hash import SHA
        from Cryptodome.PublicKey import RSA
        from Cryptodome.Signature import PKCS1_v1_5

        HAS_CRYPTO = True
    except ImportError:
        HAS_CRYPTO = False

if not HAS_M2 and not HAS_CRYPTO:
    try:
        # let this be imported, if possible
        from Crypto import Random  # nosec
        from Crypto.Cipher import AES, PKCS1_OAEP  # nosec
        from Crypto.Cipher import PKCS1_v1_5 as PKCS1_v1_5_CIPHER  # nosec
        from Crypto.Hash import SHA  # nosec
        from Crypto.PublicKey import RSA  # nosec
        from Crypto.Signature import PKCS1_v1_5  # nosec

        HAS_CRYPTO = True
    except ImportError:
        HAS_CRYPTO = False

log = logging.getLogger(__name__)


def legacy_gen_keys(keydir, keyname, keysize, user=None, passphrase=None):
    """
    Generate a RSA public keypair for use with salt

    :param str keydir: The directory to write the keypair to
    :param str keyname: The type of salt server for whom this key should be written. (i.e. 'master' or 'minion')
    :param int keysize: The number of bits in the key
    :param str user: The user on the system who should own this keypair
    :param str passphrase: The passphrase which should be used to encrypt the private key

    :rtype: str
    :return: Path on the filesystem to the RSA private key
    """
    base = os.path.join(keydir, keyname)
    priv = f"{base}.pem"
    pub = f"{base}.pub"

    # gen = rsa.generate_private_key(e, keysize)
    if HAS_M2:
        gen = RSA.gen_key(keysize, 65537, lambda: None)
    else:
        gen = RSA.generate(bits=keysize, e=65537)

    if os.path.isfile(priv):
        # Between first checking and the generation another process has made
        # a key! Use the winner's key
        return priv

    # Do not try writing anything, if directory has no permissions.
    if not os.access(keydir, os.W_OK):
        raise OSError(
            'Write access denied to "{}" for user "{}".'.format(
                os.path.abspath(keydir), getpass.getuser()
            )
        )

    with salt.utils.files.set_umask(0o277):
        if HAS_M2:
            # if passphrase is empty or None use no cipher
            if not passphrase:
                gen.save_pem(priv, cipher=None)
            else:
                gen.save_pem(
                    priv,
                    cipher="des_ede3_cbc",
                    callback=lambda x: salt.utils.stringutils.to_bytes(passphrase),
                )
        else:
            with salt.utils.files.fopen(priv, "wb+") as f:
                f.write(gen.exportKey("PEM", passphrase))
    if HAS_M2:
        gen.save_pub_key(pub)
    else:
        with salt.utils.files.fopen(pub, "wb+") as f:
            f.write(gen.publickey().exportKey("PEM"))
    os.chmod(priv, 0o400)
    if user:
        try:
            import pwd

            uid = pwd.getpwnam(user).pw_uid
            os.chown(priv, uid, -1)
            os.chown(pub, uid, -1)
        except (KeyError, ImportError, OSError):
            # The specified user was not found, allow the backup systems to
            # report the error
            pass
    return priv


class LegacyPrivateKey:

    def __init__(self, path, passphrase=None):
        if HAS_M2:
            self.key = RSA.load_key(path, lambda x: bytes(passphrase))
        else:
            with salt.utils.files.fopen(path) as f:
                self.key = RSA.importKey(f.read(), passphrase)

    def encrypt(self, data):
        if HAS_M2:
            return self.key.private_encrypt(data, salt.utils.rsax931.RSA_X931_PADDING)
        else:
            return salt.utils.rsax931.RSAX931Signer(self.key.exportKey("PEM")).sign(
                data
            )

    def sign(self, data):
        if HAS_M2:
            md = EVP.MessageDigest("sha1")
            md.update(salt.utils.stringutils.to_bytes(data))
            digest = md.final()
            return self.key.sign(digest)
        else:
            signer = PKCS1_v1_5.new(self.key)
            return signer.sign(SHA.new(salt.utils.stringutils.to_bytes(data)))


class LegacyPublicKey:
    def __init__(self, path, _HAS_M2=HAS_M2):
        self._HAS_M2 = _HAS_M2
        if self._HAS_M2:
            with salt.utils.files.fopen(path, "rb") as f:
                data = f.read().replace(b"RSA ", b"")
            bio = BIO.MemoryBuffer(data)
            try:
                self.key = RSA.load_pub_key_bio(bio)
            except RSA.RSAError:
                raise InvalidKeyError("Encountered bad RSA public key")
        else:
            with salt.utils.files.fopen(path) as f:
                try:
                    self.key = RSA.importKey(f.read())
                except (ValueError, IndexError, TypeError):
                    raise InvalidKeyError("Encountered bad RSA public key")

    def encrypt(self, data):
        bdata = salt.utils.stringutils.to_bytes(data)
        if self._HAS_M2:
            return self.key.public_encrypt(bdata, RSA.pkcs1_oaep_padding)
        else:
            return PKCS1_OAEP.new(self.key).encrypt(bdata)

    def verify(self, data, signature):
        if self._HAS_M2:
            md = EVP.MessageDigest("sha1")
            md.update(salt.utils.stringutils.to_bytes(data))
            digest = md.final()
            try:
                return self.key.verify(digest, signature)
            except RSA.RSAError as exc:
                log.debug("Signature verification failed: %s", exc.args[0])
                return False
        else:
            verifier = PKCS1_v1_5.new(self.key)
            return verifier.verify(
                SHA.new(salt.utils.stringutils.to_bytes(data)), signature
            )

    def decrypt(self, data):
        data = salt.utils.stringutils.to_bytes(data)
        if HAS_M2:
            return self.key.public_decrypt(data, salt.utils.rsax931.RSA_X931_PADDING)
        else:
            verifier = salt.utils.rsax931.RSAX931Verifier(self.key.exportKey("PEM"))
            return verifier.verify(data)


PRIVKEY_DATA = (
    "-----BEGIN RSA PRIVATE KEY-----\n"
    "MIIEpAIBAAKCAQEA75GR6ZTv5JOv90Vq8tKhKC7YQnhDIo2hM0HVziTEk5R4UQBW\n"
    "a0CKytFMbTONY2msEDwX9iA0x7F5Lgj0X8eD4ZMsYqLzqjWMekLC8bjhxc+EuPo9\n"
    "Dygu3mJ2VgRC7XhlFpmdo5NN8J2E7B/CNB3R4hOcMMZNZdi0xLtFoTfwU61UPfFX\n"
    "14mV2laqLbvDEfQLJhUTDeFFV8EN5Z4H1ttLP3sMXJvc3EvM0JiDVj4l1TWFUHHz\n"
    "eFgCA1Im0lv8i7PFrgW7nyMfK9uDSsUmIp7k6ai4tVzwkTmV5PsriP1ju88Lo3MB\n"
    "4/sUmDv/JmlZ9YyzTO3Po8Uz3Aeq9HJWyBWHAQIDAQABAoIBAGOzBzBYZUWRGOgl\n"
    "IY8QjTT12dY/ymC05GM6gMobjxuD7FZ5d32HDLu/QrknfS3kKlFPUQGDAbQhbbb0\n"
    "zw6VL5NO9mfOPO2W/3FaG1sRgBQcerWonoSSSn8OJwVBHMFLG3a+U1Zh1UvPoiPK\n"
    "S734swIM+zFpNYivGPvOm/muF/waFf8tF/47t1cwt/JGXYQnkG/P7z0vp47Irpsb\n"
    "Yjw7vPe4BnbY6SppSxscW3KoV7GtJLFKIxAXbxsuJMF/rYe3O3w2VKJ1Sug1VDJl\n"
    "/GytwAkSUer84WwP2b07Wn4c5pCnmLslMgXCLkENgi1NnJMhYVOnckxGDZk54hqP\n"
    "9RbLnkkCgYEA/yKuWEvgdzYRYkqpzB0l9ka7Y00CV4Dha9Of6GjQi9i4VCJ/UFVr\n"
    "UlhTo5y0ZzpcDAPcoZf5CFZsD90a/BpQ3YTtdln2MMCL/Kr3QFmetkmDrt+3wYnX\n"
    "sKESfsa2nZdOATRpl1antpwyD4RzsAeOPwBiACj4fkq5iZJBSI0bxrMCgYEA8GFi\n"
    "qAjgKh81/Uai6KWTOW2kX02LEMVRrnZLQ9VPPLGid4KZDDk1/dEfxjjkcyOxX1Ux\n"
    "Klu4W8ZEdZyzPcJrfk7PdopfGOfrhWzkREK9C40H7ou/1jUecq/STPfSOmxh3Y+D\n"
    "ifMNO6z4sQAHx8VaHaxVsJ7SGR/spr0pkZL+NXsCgYEA84rIgBKWB1W+TGRXJzdf\n"
    "yHIGaCjXpm2pQMN3LmP3RrcuZWm0vBt94dHcrR5l+u/zc6iwEDTAjJvqdU4rdyEr\n"
    "tfkwr7v6TNlQB3WvpWanIPyVzfVSNFX/ZWSsAgZvxYjr9ixw6vzWBXOeOb/Gqu7b\n"
    "cvpLkjmJ0wxDhbXtyXKhZA8CgYBZyvcQb+hUs732M4mtQBSD0kohc5TsGdlOQ1AQ\n"
    "McFcmbpnzDghkclyW8jzwdLMk9uxEeDAwuxWE/UEvhlSi6qdzxC+Zifp5NBc0fVe\n"
    "7lMx2mfJGxj5CnSqQLVdHQHB4zSXkAGB6XHbBd0MOUeuvzDPfs2voVQ4IG3FR0oc\n"
    "3/znuwKBgQChZGH3McQcxmLA28aUwOVbWssfXKdDCsiJO+PEXXlL0maO3SbnFn+Q\n"
    "Tyf8oHI5cdP7AbwDSx9bUfRPjg9dKKmATBFr2bn216pjGxK0OjYOCntFTVr0psRB\n"
    "CrKg52Qrq71/2l4V2NLQZU40Dr1bN9V+Ftd9L0pvpCAEAWpIbLXGDw==\n"
    "-----END RSA PRIVATE KEY-----"
)

PUBKEY_DATA = (
    "-----BEGIN PUBLIC KEY-----\n"
    "MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEA75GR6ZTv5JOv90Vq8tKh\n"
    "KC7YQnhDIo2hM0HVziTEk5R4UQBWa0CKytFMbTONY2msEDwX9iA0x7F5Lgj0X8eD\n"
    "4ZMsYqLzqjWMekLC8bjhxc+EuPo9Dygu3mJ2VgRC7XhlFpmdo5NN8J2E7B/CNB3R\n"
    "4hOcMMZNZdi0xLtFoTfwU61UPfFX14mV2laqLbvDEfQLJhUTDeFFV8EN5Z4H1ttL\n"
    "P3sMXJvc3EvM0JiDVj4l1TWFUHHzeFgCA1Im0lv8i7PFrgW7nyMfK9uDSsUmIp7k\n"
    "6ai4tVzwkTmV5PsriP1ju88Lo3MB4/sUmDv/JmlZ9YyzTO3Po8Uz3Aeq9HJWyBWH\n"
    "AQIDAQAB\n"
    "-----END PUBLIC KEY-----"
)

MSG = b"It's me, Mario"

SIG = (
    b"\x07\xf3\xb1\xe7\xdb\x06\xf4_\xe2\xdc\xcb!F\xfb\xbex{W\x1d\xe4E"
    b"\xd3\r\xc5\x90\xca(\x05\x1d\x99\x8b\x1aug\x9f\x95>\x94\x7f\xe3+"
    b"\x12\xfa\x9c\xd4\xb8\x02]\x0e\xa5\xa3LL\xc3\xa2\x8f+\x83Z\x1b\x17"
    b'\xbfT\xd3\xc7\xfd\x0b\xf4\xd7J\xfe^\x86q"I\xa3x\xbc\xd3$\xe9M<\xe1'
    b"\x07\xad\xf2_\x9f\xfa\xf7g(~\xd8\xf5\xe7\xda-\xa3Ko\xfc.\x99\xcf"
    b"\x9b\xb9\xc1U\x97\x82'\xcb\xc6\x08\xaa\xa0\xe4\xd0\xc1+\xfc\x86"
    b'\r\xe4y\xb1#\xd3\x1dS\x96D28\xc4\xd5\r\xd4\x98\x1a44"\xd7\xc2\xb4'
    b"]\xa7\x0f\xa7Db\x85G\x8c\xd6\x94!\x8af1O\xf6g\xd7\x03\xfd\xb3\xbc"
    b"\xce\x9f\xe7\x015\xb8\x1d]AHK\xa0\x14m\xda=O\xa7\xde\xf2\xff\x9b"
    b"\x8e\x83\xc8j\x11\x1a\x98\x85\xde\xc5\x91\x07\x84!\x12^4\xcb\xa8"
    b"\x98\x8a\x8a&#\xb9(#?\x80\x15\x9eW\xb5\x12\xd1\x95S\xf2<G\xeb\xf1"
    b"\x14H\xb2\xc4>\xc3A\xed\x86x~\xcfU\xd5Q\xfe~\x10\xd2\x9b"
)

SIGNATURE = (
    b"w\xac\xfe18o\xeb\xfb\x14+\x9e\xd1\xb7\x7fe}\xec\xd6\xe1P\x9e\xab"
    b"\xb5\x07\xe0\xc1\xfd\xda#\x04Z\x8d\x7f\x0b\x1f}:~\xb2s\x860u\x02N"
    b'\xd4q"\xb7\x86*\x8f\x1f\xd0\x9d\x11\x92\xc5~\xa68\xac>\x12H\xc2%y,'
    b"\xe6\xceU\x1e\xa3?\x0c,\xf0u\xbb\xd0[g_\xdd\x8b\xb0\x95:Y\x18\xa5*"
    b"\x99\xfd\xf3K\x92\x92 ({\xd1\xff\xd9F\xc8\xd6K\x86e\xf9\xa8\xad\xb0z"
    b"\xe3\x9dD\xf5k\x8b_<\xe7\xe7\xec\xf3\"'\xd5\xd2M\xb4\xce\x1a\xe3$"
    b"\x9c\x81\xad\xf9\x11\xf6\xf5>)\xc7\xdd\x03&\xf7\x86@ks\xa6\x05\xc2"
    b"\xd0\xbd\x1a7\xfc\xde\xe6\xb0\xad!\x12#\xc86Y\xea\xc5\xe3\xe2\xb3"
    b"\xc9\xaf\xfa\x0c\xf2?\xbf\x93w\x18\x9e\x0b\xa2a\x10:M\x05\x89\xe2W.Q"
    b"\xe8;yGT\xb1\xf2\xc6A\xd2\xc4\xbeN\xb3\xcfS\xaf\x03f\xe2\xb4)\xe7\xf6"
    b'\xdbs\xd0Z}8\xa4\xd2\x1fW*\xe6\x1c"\x8b\xd0\x18w\xb9\x7f\x9e\x96\xa3'
    b"\xd9v\xf7\x833\x8e\x01"
)

TEST_KEY = (
    "-----BEGIN RSA PUBLIC KEY-----\n"
    "MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAzLtFhsvfbFDFaUgulSEX\n"
    "Gl12XriL1DT78Ef2/u8HHaSMmPie37BLWas/zaHwI6066bIyYQJ/nUCahTaoHM7L\n"
    "GlWc0wOU6zyfpihCRQHil05Y6F+olFBoZuYbFPvtp7/hJx/D7I/0n2o/c7M5i3Y2\n"
    "3sBxAYNooIQHXHUmPQW6C9iu95ylZDW8JQzYy/EI4vCC8yQMdTK8jK1FQV0Sbwny\n"
    "qcMxSyAWDoFbnhh2P2TnO8HOWuUOaXR8ZHOJzVcDl+a6ew+medW090x3K5O1f80D\n"
    "+WjgnG6b2HG7VQpOCfM2GALD/FrxicPilvZ38X1aLhJuwjmVE4LAAv8DVNJXohaO\n"
    "WQIDAQAB\n"
    "-----END RSA PUBLIC KEY-----\n"
)

PRIV_KEY = """
-----BEGIN RSA PRIVATE KEY-----
MIIEogIBAAKCAQEAoAsMPt+4kuIG6vKyw9r3+OuZrVBee/2vDdVetW+Js5dTlgrJ
aghWWn3doGmKlEjqh7E4UTa+t2Jd6w8RSLnyHNJ/HpVhMG0M07MF6FMfILtDrrt8
ZX7eDVt8sx5gCEpYI+XG8Y07Ga9i3Hiczt+fu6HYwu96HggmG2pqkOrn3iGfqBvV
YVFJzSZYe7e4c1PeEs0xYcrA4k+apyGsMtpef8vRUrNicRLc7dAcvfhtgt2DXEZ2
d72t/CR4ygtUvPXzisaTPW0G7OWAheCloqvTIIPQIjR8htFxGTz02STVXfnhnJ0Z
k8KhqKF2v1SQvIYxsZU7jaDgl5i3zpeh58cYOwIDAQABAoIBABZUJEO7Y91+UnfC
H6XKrZEZkcnH7j6/UIaOD9YhdyVKxhsnax1zh1S9vceNIgv5NltzIsfV6vrb6v2K
Dx/F7Z0O0zR5o+MlO8ZncjoNKskex10gBEWG00Uqz/WPlddiQ/TSMJTv3uCBAzp+
S2Zjdb4wYPUlgzSgb2ygxrhsRahMcSMG9PoX6klxMXFKMD1JxiY8QfAHahPzQXy9
F7COZ0fCVo6BE+MqNuQ8tZeIxu8mOULQCCkLFwXmkz1FpfK/kNRmhIyhxwvCS+z4
JuErW3uXfE64RLERiLp1bSxlDdpvRO2R41HAoNELTsKXJOEt4JANRHm/CeyA5wsh
NpscufUCgYEAxhgPfcMDy2v3nL6KtkgYjdcOyRvsAF50QRbEa8ldO+87IoMDD/Oe
osFERJ5hhyyEO78QnaLVegnykiw5DWEF02RKMhD/4XU+1UYVhY0wJjKQIBadsufB
2dnaKjvwzUhPh5BrBqNHl/FXwNCRDiYqXa79eWCPC9OFbZcUWWq70s8CgYEAztOI
61zRfmXJ7f70GgYbHg+GA7IrsAcsGRITsFR82Ho0lqdFFCxz7oK8QfL6bwMCGKyk
nzk+twh6hhj5UNp18KN8wktlo02zTgzgemHwaLa2cd6xKgmAyuPiTgcgnzt5LVNG
FOjIWkLwSlpkDTl7ZzY2QSy7t+mq5d750fpIrtUCgYBWXZUbcpPL88WgDB7z/Bjg
dlvW6JqLSqMK4b8/cyp4AARbNp12LfQC55o5BIhm48y/M70tzRmfvIiKnEc/gwaE
NJx4mZrGFFURrR2i/Xx5mt/lbZbRsmN89JM+iKWjCpzJ8PgIi9Wh9DIbOZOUhKVB
9RJEAgo70LvCnPTdS0CaVwKBgDJW3BllAvw/rBFIH4OB/vGnF5gosmdqp3oGo1Ik
jipmPAx6895AH4tquIVYrUl9svHsezjhxvjnkGK5C115foEuWXw0u60uiTiy+6Pt
2IS0C93VNMulenpnUrppE7CN2iWFAiaura0CY9fE/lsVpYpucHAWgi32Kok+ZxGL
WEttAoGAN9Ehsz4LeQxEj3x8wVeEMHF6OsznpwYsI2oVh6VxpS4AjgKYqeLVcnNi
TlZFsuQcqgod8OgzA91tdB+Rp86NygmWD5WzeKXpCOg9uA+y/YL+0sgZZHsuvbK6
PllUgXdYxqClk/hdBFB7v9AQoaj7K9Ga22v32msftYDQRJ94xOI=
-----END RSA PRIVATE KEY-----
"""


PUB_KEY = """
-----BEGIN PUBLIC KEY-----
MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAoAsMPt+4kuIG6vKyw9r3
+OuZrVBee/2vDdVetW+Js5dTlgrJaghWWn3doGmKlEjqh7E4UTa+t2Jd6w8RSLny
HNJ/HpVhMG0M07MF6FMfILtDrrt8ZX7eDVt8sx5gCEpYI+XG8Y07Ga9i3Hiczt+f
u6HYwu96HggmG2pqkOrn3iGfqBvVYVFJzSZYe7e4c1PeEs0xYcrA4k+apyGsMtpe
f8vRUrNicRLc7dAcvfhtgt2DXEZ2d72t/CR4ygtUvPXzisaTPW0G7OWAheCloqvT
IIPQIjR8htFxGTz02STVXfnhnJ0Zk8KhqKF2v1SQvIYxsZU7jaDgl5i3zpeh58cY
OwIDAQAB
-----END PUBLIC KEY-----
"""

PRIV_KEY2 = """
-----BEGIN RSA PRIVATE KEY-----
MIIEogIBAAKCAQEAp+8cTxguO6Vg+YO92VfHgNld3Zy8aM3JbZvpJcjTnis+YFJ7
Zlkcc647yPRRwY9nYBNywahnt5kIeuT1rTvTsMBZWvmUoEVUj1Xg8XXQkBvb9Ozy
Gqy/G/p8KDDpzMP/U+XCnUeHiXTZrgnqgBIc2cKeCVvWFqDi0GRFGzyaXLaX3PPm
M7DJ0MIPL1qgmcDq6+7Ze0gJ9SrDYFAeLmbuT1OqDfufXWQl/82JXeiwU2cOpqWq
7n5fvPOWim7l1tzQ+dSiMRRm0xa6uNexCJww3oJSwvMbAmgzvOhqqhlqv+K7u0u7
FrFFojESsL36Gq4GBrISnvu2tk7u4GGNTYYQbQIDAQABAoIBAADrqWDQnd5DVZEA
lR+WINiWuHJAy/KaIC7K4kAMBgbxrz2ZbiY9Ok/zBk5fcnxIZDVtXd1sZicmPlro
GuWodIxdPZAnWpZ3UtOXUayZK/vCP1YsH1agmEqXuKsCu6Fc+K8VzReOHxLUkmXn
FYM+tixGahXcjEOi/aNNTWitEB6OemRM1UeLJFzRcfyXiqzHpHCIZwBpTUAsmzcG
QiVDkMTKubwo/m+PVXburX2CGibUydctgbrYIc7EJvyx/cpRiPZXo1PhHQWdu4Y1
SOaC66WLsP/wqvtHo58JQ6EN/gjSsbAgGGVkZ1xMo66nR+pLpR27coS7o03xCks6
DY/0mukCgYEAuLIGgBnqoh7YsOBLd/Bc1UTfDMxJhNseo+hZemtkSXz2Jn51322F
Zw/FVN4ArXgluH+XsOhvG/MFFpojwZSrb0Qq5b1MRdo9qycq8lGqNtlN1WHqosDQ
zW29kpL0tlRrSDpww3wRESsN9rH5XIrJ1b3ZXuO7asR+KBVQMy/+NcUCgYEA6MSC
c+fywltKPgmPl5j0DPoDe5SXE/6JQy7w/vVGrGfWGf/zEJmhzS2R+CcfTTEqaT0T
Yw8+XbFgKAqsxwtE9MUXLTVLI3sSUyE4g7blCYscOqhZ8ItCUKDXWkSpt++rG0Um
1+cEJP/0oCazG6MWqvBC4NpQ1nzh46QpjWqMwokCgYAKDLXJ1p8rvx3vUeUJW6zR
dfPlEGCXuAyMwqHLxXgpf4EtSwhC5gSyPOtx2LqUtcrnpRmt6JfTH4ARYMW9TMef
QEhNQ+WYj213mKP/l235mg1gJPnNbUxvQR9lkFV8bk+AGJ32JRQQqRUTbU+yN2MQ
HEptnVqfTp3GtJIultfwOQKBgG+RyYmu8wBP650izg33BXu21raEeYne5oIqXN+I
R5DZ0JjzwtkBGroTDrVoYyuH1nFNEh7YLqeQHqvyufBKKYo9cid8NQDTu+vWr5UK
tGvHnwdKrJmM1oN5JOAiq0r7+QMAOWchVy449VNSWWV03aeftB685iR5BXkstbIQ
EVopAoGAfcGBTAhmceK/4Q83H/FXBWy0PAa1kZGg/q8+Z0KY76AqyxOVl0/CU/rB
3tO3sKhaMTHPME/MiQjQQGoaK1JgPY6JHYvly2KomrJ8QTugqNGyMzdVJkXAK2AM
GAwC8ivAkHf8CHrHa1W7l8t2IqBjW1aRt7mOW92nfG88Hck0Mbo=
-----END RSA PRIVATE KEY-----
"""


PUB_KEY2 = """
-----BEGIN PUBLIC KEY-----
MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAp+8cTxguO6Vg+YO92VfH
gNld3Zy8aM3JbZvpJcjTnis+YFJ7Zlkcc647yPRRwY9nYBNywahnt5kIeuT1rTvT
sMBZWvmUoEVUj1Xg8XXQkBvb9OzyGqy/G/p8KDDpzMP/U+XCnUeHiXTZrgnqgBIc
2cKeCVvWFqDi0GRFGzyaXLaX3PPmM7DJ0MIPL1qgmcDq6+7Ze0gJ9SrDYFAeLmbu
T1OqDfufXWQl/82JXeiwU2cOpqWq7n5fvPOWim7l1tzQ+dSiMRRm0xa6uNexCJww
3oJSwvMbAmgzvOhqqhlqv+K7u0u7FrFFojESsL36Gq4GBrISnvu2tk7u4GGNTYYQ
bQIDAQAB
-----END PUBLIC KEY-----
"""



"""
Integration tests for the alternatives state module
"""

import os

import pytest

from tests.support.case import ModuleCase
from tests.support.mixins import SaltReturnAssertsMixin
from tests.support.runtests import RUNTIME_VARS

NO_ALTERNATIVES = False
if not os.path.exists("/etc/alternatives"):
    NO_ALTERNATIVES = True


@pytest.mark.skipif(
    NO_ALTERNATIVES, reason="/etc/alternatives does not exist on the system"
)
class AlterantivesStateTest(ModuleCase, SaltReturnAssertsMixin):
    @pytest.mark.destructive_test
    @pytest.mark.slow_test
    def test_install_set_and_remove(self):
        ret = self.run_state(
            "alternatives.set", name="alt-test", path=RUNTIME_VARS.SHELL_TRUE_PATH
        )
        self.assertSaltFalseReturn(ret)

        ret = self.run_state(
            "alternatives.install",
            name="alt-test",
            link="/usr/local/bin/alt-test",
            path=RUNTIME_VARS.SHELL_TRUE_PATH,
            priority=50,
        )
        self.assertSaltTrueReturn(ret)
        self.assertSaltStateChangesEqual(
            ret, RUNTIME_VARS.SHELL_TRUE_PATH, keys=["path"]
        )

        ret = self.run_state(
            "alternatives.install",
            name="alt-test",
            link="/usr/local/bin/alt-test",
            path=RUNTIME_VARS.SHELL_TRUE_PATH,
            priority=50,
        )
        self.assertSaltTrueReturn(ret)
        self.assertSaltStateChangesEqual(ret, {})

        ret = self.run_state(
            "alternatives.install",
            name="alt-test",
            link="/usr/local/bin/alt-test",
            path=RUNTIME_VARS.SHELL_FALSE_PATH,
            priority=90,
        )
        self.assertSaltTrueReturn(ret)
        self.assertSaltStateChangesEqual(
            ret, RUNTIME_VARS.SHELL_FALSE_PATH, keys=["path"]
        )

        ret = self.run_state(
            "alternatives.set", name="alt-test", path=RUNTIME_VARS.SHELL_FALSE_PATH
        )
        self.assertSaltTrueReturn(ret)
        self.assertSaltStateChangesEqual(ret, {})

        ret = self.run_state(
            "alternatives.set", name="alt-test", path=RUNTIME_VARS.SHELL_TRUE_PATH
        )
        self.assertSaltTrueReturn(ret)
        self.assertSaltStateChangesEqual(
            ret, RUNTIME_VARS.SHELL_TRUE_PATH, keys=["path"]
        )

        ret = self.run_state(
            "alternatives.set", name="alt-test", path=RUNTIME_VARS.SHELL_TRUE_PATH
        )
        self.assertSaltTrueReturn(ret)
        self.assertSaltStateChangesEqual(ret, {})

        ret = self.run_state(
            "alternatives.remove", name="alt-test", path=RUNTIME_VARS.SHELL_TRUE_PATH
        )
        self.assertSaltTrueReturn(ret)

        ret = self.run_state(
            "alternatives.remove", name="alt-test", path=RUNTIME_VARS.SHELL_FALSE_PATH
        )
        self.assertSaltTrueReturn(ret)



"""
The client module is used to create a client connection to the publisher
The data structure needs to be:
    {'enc': 'clear',
     'load': {'fun': '<mod.callable>',
              'arg':, ('arg1', 'arg2', ...),
              'tgt': '<glob or id>',
              'key': '<read in the key file>'}
"""

import logging

# The components here are simple, and they need to be and stay simple, we
# want a client to have 3 external concerns, and maybe a forth configurable
# option.
# The concerns are:
# 1. Who executes the command?
# 2. What is the function being run?
# 3. What arguments need to be passed to the function?
# 4. How long do we wait for all of the replies?
import os
import random
import sys
import time
from datetime import datetime

import salt.cache
import salt.channel.client
import salt.config
import salt.defaults.exitcodes
import salt.ext.tornado.gen
import salt.loader
import salt.payload
import salt.syspaths as syspaths
import salt.utils.args
import salt.utils.event
import salt.utils.files
import salt.utils.jid
import salt.utils.minions
import salt.utils.network
import salt.utils.platform
import salt.utils.stringutils
import salt.utils.user
import salt.utils.verify
from salt.exceptions import (
    AuthenticationError,
    AuthorizationError,
    EauthAuthenticationError,
    PublishError,
    SaltClientError,
    SaltInvocationError,
    SaltReqTimeoutError,
)

HAS_RANGE = False
try:
    import seco.range

    HAS_RANGE = True
except ImportError:
    pass


log = logging.getLogger(__name__)


def get_local_client(
    c_path=os.path.join(syspaths.CONFIG_DIR, "master"),
    mopts=None,
    skip_perm_errors=False,
    io_loop=None,
    auto_reconnect=False,
    listen=False,
):
    """
    .. versionadded:: 2014.7.0

    Read in the config and return the correct LocalClient object based on
    the configured transport

    :param str c_path: Path of config file to use for opts.

                       The default value is None.

    :param bool mopts: When provided the local client will use this dictionary of
                       options insead of loading a config file from the value
                       of c_path.

                       The default value is None.

    :param str skip_perm_errors: Ignore permissions errors while loading keys.

                                 The default value is False.

    :param IOLoop io_loop: io_loop used for events.
                           Pass in an io_loop if you want asynchronous
                           operation for obtaining events. Eg use of
                           set_event_handler() API. Otherwise, operation
                           will be synchronous.

    :param bool keep_loop: Do not destroy the event loop when closing the event
                           subsriber.

    :param bool auto_reconnect: When True the event subscriber will reconnect
                                automatically if a disconnect error is raised.

    .. versionadded:: 3004
    :param bool listen: Listen for events indefinitly. When option is set the
                        LocalClient object will listen for events until it's
                        destroy method is called.

                        The default value is False.
    """
    if mopts:
        opts = mopts
    else:
        # Late import to prevent circular import
        import salt.config

        opts = salt.config.client_config(c_path)

    # TODO: AIO core is separate from transport
    return LocalClient(
        mopts=opts,
        skip_perm_errors=skip_perm_errors,
        io_loop=io_loop,
        auto_reconnect=auto_reconnect,
        listen=listen,
    )


class LocalClient:
    """
    The interface used by the :command:`salt` CLI tool on the Salt Master

    ``LocalClient`` is used to send a command to Salt minions to execute
    :ref:`execution modules <all-salt.modules>` and return the results to the
    Salt Master.

    Importing and using ``LocalClient`` must be done on the same machine as the
    Salt Master and it must be done using the same user that the Salt Master is
    running as. (Unless :conf_master:`external_auth` is configured and
    authentication credentials are included in the execution).

    .. note::
        The LocalClient uses a Tornado IOLoop, this can create issues when
        using the LocalClient inside an existing IOLoop. If creating the
        LocalClient in partnership with another IOLoop either create the
        IOLoop before creating the LocalClient, or when creating the IOLoop
        use ioloop.current() which will return the ioloop created by
        LocalClient.

    .. code-block:: python

        import salt.client

        local = salt.client.LocalClient()
        local.cmd('*', 'test.fib', [10])

    """

    def __init__(
        self,
        c_path=os.path.join(syspaths.CONFIG_DIR, "master"),
        mopts=None,
        skip_perm_errors=False,
        io_loop=None,
        keep_loop=False,
        auto_reconnect=False,
        listen=False,
    ):
        """
        :param str c_path: Path of config file to use for opts.

                           The default value is None.

        :param bool mopts: When provided the local client will use this dictionary of
                           options insead of loading a config file from the value
                           of c_path.

                           The default value is None.

        :param str skip_perm_errors: Ignore permissions errors while loading keys.

                                     The default value is False.

        :param IOLoop io_loop: io_loop used for events.
                               Pass in an io_loop if you want asynchronous
                               operation for obtaining events. Eg use of
                               set_event_handler() API. Otherwise, operation
                               will be synchronous.

        :param bool keep_loop: Do not destroy the event loop when closing the event
                               subsriber.

        :param bool auto_reconnect: When True the event subscriber will reconnect
                                    automatically if a disconnect error is raised.

        .. versionadded:: 3004
        :param bool listen: Listen for events indefinitly. When option is set the
                            LocalClient object will listen for events until it's
                            destroy method is called.

                            The default value is False.
        """
        if mopts:
            self.opts = mopts
        else:
            if os.path.isdir(c_path):
                log.warning(
                    "%s expects a file path not a directory path(%s) to "
                    "its 'c_path' keyword argument",
                    self.__class__.__name__,
                    c_path,
                )
            self.opts = salt.config.client_config(c_path)
        self.salt_user = salt.utils.user.get_specific_user()
        self.skip_perm_errors = skip_perm_errors
        self.key = self.__read_master_key()
        self.auto_reconnect = auto_reconnect
        self.listen = listen
        self.event = salt.utils.event.get_event(
            "master",
            self.opts["sock_dir"],
            opts=self.opts,
            listen=self.listen,
            io_loop=io_loop,
            keep_loop=keep_loop,
        )
        self.utils = salt.loader.utils(self.opts)
        self.functions = salt.loader.minion_mods(self.opts, utils=self.utils)
        self.returners = salt.loader.returners(self.opts, self.functions)

    def __read_master_key(self):
        """
        Read in the rotating master authentication key
        """
        key_user = self.salt_user
        if key_user == "root":
            if self.opts.get("user", "root") != "root":
                key_user = self.opts.get("user", "root")
        if key_user.startswith("sudo_"):
            key_user = self.opts.get("user", "root")
        if salt.utils.platform.is_windows():
            # The username may contain '\' if it is in Windows
            # 'DOMAIN\username' format. Fix this for the keyfile path.
            key_user = key_user.replace("\\", "_")
        keyfile = os.path.join(self.opts["cachedir"], f".{key_user}_key")
        try:
            # Make sure all key parent directories are accessible
            salt.utils.verify.check_path_traversal(
                self.opts["cachedir"], key_user, self.skip_perm_errors
            )
            with salt.utils.files.fopen(keyfile, "r") as key:
                return salt.utils.stringutils.to_unicode(key.read())
        except (OSError, SaltClientError):
            # Fall back to eauth
            return ""

    def _convert_range_to_list(self, tgt):
        """
        convert a seco.range range into a list target
        """
        range_ = seco.range.Range(self.opts["range_server"])
        try:
            return range_.expand(tgt)
        except seco.range.RangeException as err:
            print(f"Range server exception: {err}")
            return []

    def _get_timeout(self, timeout):
        """
        Return the timeout to use
        """
        if timeout is None:
            return self.opts["timeout"]
        if isinstance(timeout, int):
            return timeout
        if isinstance(timeout, str):
            try:
                return int(timeout)
            except ValueError:
                return self.opts["timeout"]
        # Looks like the timeout is invalid, use config
        return self.opts["timeout"]

    def gather_job_info(self, jid, tgt, tgt_type, listen=True, **kwargs):
        """
        Return the information about a given job
        """
        log.debug("Checking whether jid %s is still running", jid)
        timeout = int(kwargs.get("gather_job_timeout", self.opts["gather_job_timeout"]))

        pub_data = self.run_job(
            tgt,
            "saltutil.find_job",
            arg=[jid],
            tgt_type=tgt_type,
            timeout=timeout,
            listen=listen,
            **kwargs,
        )

        if "jid" in pub_data:
            self.event.subscribe(pub_data["jid"])

        return pub_data

    def _check_pub_data(self, pub_data, listen=True):
        """
        Common checks on the pub_data data structure returned from running pub
        """
        if pub_data == "":
            # Failed to authenticate, this could be a bunch of things
            raise EauthAuthenticationError(
                "Failed to authenticate! This is most likely because this "
                "user is not permitted to execute commands, but there is a "
                "small possibility that a disk error occurred (check "
                "disk/inode usage)."
            )

        # Failed to connect to the master and send the pub
        if "error" in pub_data:
            print(pub_data["error"])
            log.debug("_check_pub_data() error: %s", pub_data["error"])
            return {}
        elif "jid" not in pub_data:
            return {}
        if pub_data["jid"] == "0":
            print("Failed to connect to the Master, is the Salt Master running?")
            return {}

        # If we order masters (via a syndic), don't short circuit if no minions
        # are found
        if not self.opts.get("order_masters"):
            # Check for no minions
            if not pub_data["minions"]:
                print(
                    "No minions matched the target. "
                    "No command was sent, no jid was assigned."
                )
                return {}

        # don't install event subscription listeners when the request is asynchronous
        # and doesn't care. this is important as it will create event leaks otherwise
        if not listen:
            return pub_data

        if self.opts.get("order_masters"):
            self.event.subscribe("syndic/.*/{}".format(pub_data["jid"]), "regex")

        self.event.subscribe("salt/job/{}".format(pub_data["jid"]))

        return pub_data

    def run_job(
        self,
        tgt,
        fun,
        arg=(),
        tgt_type="glob",
        ret="",
        timeout=None,
        jid="",
        kwarg=None,
        listen=False,
        **kwargs,
    ):
        """
        Asynchronously send a command to connected minions

        Prep the job directory and publish a command to any targeted minions.

        :return: A dictionary of (validated) ``pub_data`` or an empty
            dictionary on failure. The ``pub_data`` contains the job ID and a
            list of all minions that are expected to return data.

        .. code-block:: python

            >>> local.run_job('*', 'test.sleep', [300])
            {'jid': '20131219215650131543', 'minions': ['jerry']}
        """
        arg = salt.utils.args.condition_input(arg, kwarg)

        try:
            pub_data = self.pub(
                tgt,
                fun,
                arg,
                tgt_type,
                ret,
                jid=jid,
                timeout=self._get_timeout(timeout),
                listen=listen,
                **kwargs,
            )
        except SaltClientError:
            # Re-raise error with specific message
            raise SaltClientError(
                "The salt master could not be contacted. Is master running?"
            )
        except AuthenticationError as err:
            raise
        except AuthorizationError as err:
            raise
        except Exception as general_exception:  # pylint: disable=broad-except
            # Convert to generic client error and pass along message
            raise SaltClientError(general_exception)

        return self._check_pub_data(pub_data, listen=listen)

    def gather_minions(self, tgt, expr_form):
        _res = salt.utils.minions.CkMinions(self.opts).check_minions(
            tgt, tgt_type=expr_form
        )
        return _res["minions"]

    @salt.ext.tornado.gen.coroutine
    def run_job_async(
        self,
        tgt,
        fun,
        arg=(),
        tgt_type="glob",
        ret="",
        timeout=None,
        jid="",
        kwarg=None,
        listen=True,
        io_loop=None,
        **kwargs,
    ):
        """
        Asynchronously send a command to connected minions

        Prep the job directory and publish a command to any targeted minions.

        :return: A dictionary of (validated) ``pub_data`` or an empty
            dictionary on failure. The ``pub_data`` contains the job ID and a
            list of all minions that are expected to return data.

        .. code-block:: python

            >>> local.run_job_async('*', 'test.sleep', [300])
            {'jid': '20131219215650131543', 'minions': ['jerry']}
        """
        arg = salt.utils.args.condition_input(arg, kwarg)

        try:
            pub_data = yield self.pub_async(
                tgt,
                fun,
                arg,
                tgt_type,
                ret,
                jid=jid,
                timeout=self._get_timeout(timeout),
                io_loop=io_loop,
                listen=listen,
                **kwargs,
            )
        except SaltClientError:
            # Re-raise error with specific message
            raise SaltClientError(
                "The salt master could not be contacted. Is master running?"
            )
        except AuthenticationError as err:
            raise AuthenticationError(err)
        except AuthorizationError as err:
            raise AuthorizationError(err)
        except Exception as general_exception:  # pylint: disable=broad-except
            # Convert to generic client error and pass along message
            raise SaltClientError(general_exception)

        raise salt.ext.tornado.gen.Return(self._check_pub_data(pub_data, listen=listen))

    def cmd_async(
        self, tgt, fun, arg=(), tgt_type="glob", ret="", jid="", kwarg=None, **kwargs
    ):
        """
        Asynchronously send a command to connected minions

        The function signature is the same as :py:meth:`cmd` with the
        following exceptions.

        :returns: A job ID or 0 on failure.

        .. code-block:: python

            >>> local.cmd_async('*', 'test.sleep', [300])
            '20131219215921857715'
        """
        pub_data = self.run_job(
            tgt, fun, arg, tgt_type, ret, jid=jid, kwarg=kwarg, listen=False, **kwargs
        )
        try:
            return pub_data["jid"]
        except KeyError:
            return 0

    def cmd_subset(
        self,
        tgt,
        fun,
        arg=(),
        tgt_type="glob",
        ret="",
        kwarg=None,
        subset=3,
        cli=False,
        progress=False,
        full_return=False,
        **kwargs,
    ):
        """
        Execute a command on a random subset of the targeted systems

        The function signature is the same as :py:meth:`cmd` with the
        following exceptions.

        :param subset: The number of systems to execute on
        :param cli: When this is set to True, a generator is returned,
                    otherwise a dictionary of the minion returns is returned

        .. code-block:: python

            >>> SLC.cmd_subset('*', 'test.ping', subset=1)
            {'jerry': True}
        """

        # Support legacy parameter name:
        subset = kwargs.pop("sub", subset)

        minion_ret = self.cmd(tgt, "sys.list_functions", tgt_type=tgt_type, **kwargs)
        minions = list(minion_ret)
        random.shuffle(minions)
        f_tgt = []
        for minion in minions:
            if fun in minion_ret[minion]:
                f_tgt.append(minion)
            if len(f_tgt) >= subset:
                break
        func = self.cmd
        if cli:
            func = self.cmd_cli
        return func(
            f_tgt,
            fun,
            arg,
            tgt_type="list",
            ret=ret,
            kwarg=kwarg,
            progress=progress,
            full_return=full_return,
            **kwargs,
        )

    def cmd_batch(
        self,
        tgt,
        fun,
        arg=(),
        tgt_type="glob",
        ret="",
        kwarg=None,
        batch="10%",
        **kwargs,
    ):
        """
        Iteratively execute a command on subsets of minions at a time

        The function signature is the same as :py:meth:`cmd` with the
        following exceptions.

        :param batch: The batch identifier of systems to execute on

        :returns: A generator of minion returns

        .. code-block:: python

            >>> returns = local.cmd_batch('*', 'state.highstate', batch='10%')
            >>> for ret in returns:
            ...     print(ret)
            {'jerry': {...}}
            {'dave': {...}}
            {'stewart': {...}}
        """
        # We need to re-import salt.utils.args here
        # even though it has already been imported.
        # when cmd_batch is called via the NetAPI
        # the module is unavailable.
        # Late import - not used anywhere else in this file
        import salt.cli.batch
        import salt.utils.args

        arg = salt.utils.args.condition_input(arg, kwarg)
        opts = {
            "tgt": tgt,
            "fun": fun,
            "arg": arg,
            "tgt_type": tgt_type,
            "ret": ret,
            "batch": batch,
            "failhard": kwargs.get("failhard", self.opts.get("failhard", False)),
            "raw": kwargs.get("raw", False),
        }

        if "timeout" in kwargs:
            opts["timeout"] = kwargs["timeout"]
        if "gather_job_timeout" in kwargs:
            opts["gather_job_timeout"] = kwargs["gather_job_timeout"]
        if "batch_wait" in kwargs:
            opts["batch_wait"] = int(kwargs["batch_wait"])

        eauth = {}
        if "eauth" in kwargs:
            eauth["eauth"] = kwargs.pop("eauth")
        if "username" in kwargs:
            eauth["username"] = kwargs.pop("username")
        if "password" in kwargs:
            eauth["password"] = kwargs.pop("password")
        if "token" in kwargs:
            eauth["token"] = kwargs.pop("token")

        for key, val in self.opts.items():
            if key not in opts:
                opts[key] = val
        batch = salt.cli.batch.Batch(opts, eauth=eauth, quiet=True)
        for ret, _ in batch.run():
            yield ret

    def cmd(
        self,
        tgt,
        fun,
        arg=(),
        timeout=None,
        tgt_type="glob",
        ret="",
        jid="",
        full_return=False,
        kwarg=None,
        **kwargs,
    ):
        """
        Synchronously execute a command on targeted minions

        The cmd method will execute and wait for the timeout period for all
        minions to reply, then it will return all minion data at once.

        .. code-block:: python

            >>> import salt.client
            >>> local = salt.client.LocalClient()
            >>> local.cmd('*', 'cmd.run', ['whoami'])
            {'jerry': 'root'}

        With extra keyword arguments for the command function to be run:

        .. code-block:: python

            local.cmd('*', 'test.arg', ['arg1', 'arg2'], kwarg={'foo': 'bar'})

        Compound commands can be used for multiple executions in a single
        publish. Function names and function arguments are provided in separate
        lists but the index values must correlate and an empty list must be
        used if no arguments are required.

        .. code-block:: python

            >>> local.cmd('*', [
                    'grains.items',
                    'sys.doc',
                    'cmd.run',
                ],
                [
                    [],
                    [],
                    ['uptime'],
                ])

        :param tgt: Which minions to target for the execution. Default is shell
            glob. Modified by the ``tgt_type`` option.
        :type tgt: string or list

        :param fun: The module and function to call on the specified minions of
            the form ``module.function``. For example ``test.ping`` or
            ``grains.items``.

            Compound commands
                Multiple functions may be called in a single publish by
                passing a list of commands. This can dramatically lower
                overhead and speed up the application communicating with Salt.

                This requires that the ``arg`` param is a list of lists. The
                ``fun`` list and the ``arg`` list must correlate by index
                meaning a function that does not take arguments must still have
                a corresponding empty list at the expected index.
        :type fun: string or list of strings

        :param arg: A list of arguments to pass to the remote function. If the
            function takes no arguments ``arg`` may be omitted except when
            executing a compound command.
        :type arg: list or list-of-lists

        :param timeout: Seconds to wait after the last minion returns but
            before all minions return.

        :param tgt_type: The type of ``tgt``. Allowed values:

            * ``glob`` - Bash glob completion - Default
            * ``pcre`` - Perl style regular expression
            * ``list`` - Python list of hosts
            * ``grain`` - Match based on a grain comparison
            * ``grain_pcre`` - Grain comparison with a regex
            * ``pillar`` - Pillar data comparison
            * ``pillar_pcre`` - Pillar data comparison with a regex
            * ``nodegroup`` - Match on nodegroup
            * ``range`` - Use a Range server for matching
            * ``compound`` - Pass a compound match string
            * ``ipcidr`` - Match based on Subnet (CIDR notation) or IPv4 address.

            .. versionchanged:: 2017.7.0
                Renamed from ``expr_form`` to ``tgt_type``

        :param ret: The returner to use. The value passed can be single
            returner, or a comma delimited list of returners to call in order
            on the minions

        :param kwarg: A dictionary with keyword arguments for the function.

        :param full_return: Output the job return only (default) or the full
            return including exit code and other job metadata.

        :param kwargs: Optional keyword arguments.
            Authentication credentials may be passed when using
            :conf_master:`external_auth`.

            For example: ``local.cmd('*', 'test.ping', username='saltdev',
            password='saltdev', eauth='pam')``.
            Or: ``local.cmd('*', 'test.ping',
            token='5871821ea51754fdcea8153c1c745433')``

        :returns: A dictionary with the result of the execution, keyed by
            minion ID. A compound command will return a sub-dictionary keyed by
            function name.
        """
        was_listening = self.event.cpub

        try:
            pub_data = self.run_job(
                tgt,
                fun,
                arg,
                tgt_type,
                ret,
                timeout,
                jid,
                kwarg=kwarg,
                listen=True,
                **kwargs,
            )

            if not pub_data:
                return pub_data

            ret = {}
            for fn_ret in self.get_cli_event_returns(
                pub_data["jid"],
                pub_data["minions"],
                self._get_timeout(timeout),
                tgt,
                tgt_type,
                **kwargs,
            ):

                if fn_ret:
                    for mid, data in fn_ret.items():
                        ret[mid] = data if full_return else data.get("ret", {})

            for failed in list(set(pub_data["minions"]) - set(ret)):
                ret[failed] = False
            return ret
        finally:
            if not was_listening:
                self.event.close_pub()

    def cmd_cli(
        self,
        tgt,
        fun,
        arg=(),
        timeout=None,
        tgt_type="glob",
        ret="",
        verbose=False,
        kwarg=None,
        progress=False,
        **kwargs,
    ):
        """
        Used by the :command:`salt` CLI. This method returns minion returns as
        they come back and attempts to block until all minions return.

        The function signature is the same as :py:meth:`cmd` with the
        following exceptions.

        :param verbose: Print extra information about the running command
        :returns: A generator
        """
        was_listening = self.event.cpub

        try:
            self.pub_data = self.run_job(
                tgt,
                fun,
                arg,
                tgt_type,
                ret,
                timeout,
                kwarg=kwarg,
                listen=True,
                **kwargs,
            )
            if not self.pub_data:
                yield self.pub_data
            else:
                try:
                    for fn_ret in self.get_cli_event_returns(
                        self.pub_data["jid"],
                        self.pub_data["minions"],
                        self._get_timeout(timeout),
                        tgt,
                        tgt_type,
                        verbose,
                        progress,
                        **kwargs,
                    ):

                        if not fn_ret:
                            continue

                        yield fn_ret
                except KeyboardInterrupt:
                    raise SystemExit(
                        "\n"
                        "This job's jid is: {0}\n"
                        "Exiting gracefully on Ctrl-c\n"
                        "The minions may not have all finished running and any "
                        "remaining minions will return upon completion. To look "
                        "up the return data for this job later, run the following "
                        "command:\n\n"
                        "salt-run jobs.lookup_jid {0}".format(self.pub_data["jid"])
                    )
        finally:
            if not was_listening:
                self.event.close_pub()

    def cmd_iter(
        self,
        tgt,
        fun,
        arg=(),
        timeout=None,
        tgt_type="glob",
        ret="",
        kwarg=None,
        **kwargs,
    ):
        """
        Yields the individual minion returns as they come in

        The function signature is the same as :py:meth:`cmd` with the
        following exceptions.

        Normally :py:meth:`cmd_iter` does not yield results for minions that
        are not connected. If you want it to return results for disconnected
        minions set `expect_minions=True` in `kwargs`.

        :return: A generator yielding the individual minion returns

        .. code-block:: python

            >>> ret = local.cmd_iter('*', 'test.ping')
            >>> for i in ret:
            ...     print(i)
            {'jerry': {'ret': True}}
            {'dave': {'ret': True}}
            {'stewart': {'ret': True}}
        """
        was_listening = self.event.cpub

        try:
            pub_data = self.run_job(
                tgt,
                fun,
                arg,
                tgt_type,
                ret,
                timeout,
                kwarg=kwarg,
                listen=True,
                **kwargs,
            )

            if not pub_data:
                yield pub_data
            else:
                if kwargs.get("yield_pub_data"):
                    yield pub_data
                for fn_ret in self.get_iter_returns(
                    pub_data["jid"],
                    pub_data["minions"],
                    timeout=self._get_timeout(timeout),
                    tgt=tgt,
                    tgt_type=tgt_type,
                    **kwargs,
                ):
                    if not fn_ret:
                        continue
                    yield fn_ret
                self._clean_up_subscriptions(pub_data["jid"])
        finally:
            if not was_listening:
                self.event.close_pub()

    def cmd_iter_no_block(
        self,
        tgt,
        fun,
        arg=(),
        timeout=None,
        tgt_type="glob",
        ret="",
        kwarg=None,
        show_jid=False,
        verbose=False,
        **kwargs,
    ):
        """
        Yields the individual minion returns as they come in, or None
            when no returns are available.

        The function signature is the same as :py:meth:`cmd` with the
        following exceptions.

        :returns: A generator yielding the individual minion returns, or None
            when no returns are available. This allows for actions to be
            injected in between minion returns.

        .. code-block:: python

            >>> ret = local.cmd_iter_no_block('*', 'test.ping')
            >>> for i in ret:
            ...     print(i)
            None
            {'jerry': {'ret': True}}
            {'dave': {'ret': True}}
            None
            {'stewart': {'ret': True}}
        """
        was_listening = self.event.cpub

        try:
            pub_data = self.run_job(
                tgt,
                fun,
                arg,
                tgt_type,
                ret,
                timeout,
                kwarg=kwarg,
                listen=True,
                **kwargs,
            )

            if not pub_data:
                yield pub_data
            else:
                for fn_ret in self.get_iter_returns(
                    pub_data["jid"],
                    pub_data["minions"],
                    timeout=timeout,
                    tgt=tgt,
                    tgt_type=tgt_type,
                    block=False,
                    **kwargs,
                ):
                    if fn_ret and any([show_jid, verbose]):
                        for minion in fn_ret:
                            fn_ret[minion]["jid"] = pub_data["jid"]
                    yield fn_ret

                self._clean_up_subscriptions(pub_data["jid"])
        finally:
            if not was_listening:
                self.event.close_pub()

    def cmd_full_return(
        self,
        tgt,
        fun,
        arg=(),
        timeout=None,
        tgt_type="glob",
        ret="",
        verbose=False,
        kwarg=None,
        **kwargs,
    ):
        """
        Execute a salt command and return
        """
        was_listening = self.event.cpub

        try:
            pub_data = self.run_job(
                tgt,
                fun,
                arg,
                tgt_type,
                ret,
                timeout,
                kwarg=kwarg,
                listen=True,
                **kwargs,
            )

            if not pub_data:
                return pub_data

            return self.get_cli_static_event_returns(
                pub_data["jid"], pub_data["minions"], timeout, tgt, tgt_type, verbose
            )
        finally:
            if not was_listening:
                self.event.close_pub()

    def get_cli_returns(
        self,
        jid,
        minions,
        timeout=None,
        tgt="*",
        tgt_type="glob",
        verbose=False,
        show_jid=False,
        **kwargs,
    ):
        """
        Starts a watcher looking at the return data for a specified JID

        :returns: all of the information for the JID
        """
        if verbose:
            msg = f"Executing job with jid {jid}"
            print(msg)
            print("-" * len(msg) + "\n")
        elif show_jid:
            print(f"jid: {jid}")
        if timeout is None:
            timeout = self.opts["timeout"]
        fret = {}
        # make sure the minions is a set (since we do set operations on it)
        minions = set(minions)

        found = set()
        # start this before the cache lookup-- in case new stuff comes in
        event_iter = self.get_event_iter_returns(jid, minions, timeout=timeout)

        # get the info from the cache
        ret = self.get_cache_returns(jid)
        if ret != {}:
            found.update(set(ret))
            yield ret

        # if you have all the returns, stop
        if len(found.intersection(minions)) >= len(minions):
            return

        # otherwise, get them from the event system
        for event in event_iter:
            if event != {}:
                found.update(set(event))
                yield event
            if len(found.intersection(minions)) >= len(minions):
                self._clean_up_subscriptions(jid)
                return

    # TODO: tests!!
    def get_returns_no_block(self, tag, match_type=None):
        """
        Raw function to just return events of jid excluding timeout logic

        Yield either the raw event data or None

        Pass a list of additional regular expressions as `tags_regex` to search
        the event bus for non-return data, such as minion lists returned from
        syndics.
        """

        while True:
            raw = self.event.get_event(
                wait=0.01,
                tag=tag,
                match_type=match_type,
                full=True,
                no_block=True,
                auto_reconnect=self.auto_reconnect,
            )
            yield raw

    def returns_for_job(self, jid):
        return self.returners["{}.get_load".format(self.opts["master_job_cache"])](jid)

    def get_iter_returns(
        self,
        jid,
        minions,
        timeout=None,
        tgt="*",
        tgt_type="glob",
        expect_minions=False,
        block=True,
        **kwargs,
    ):
        """
        Watch the event system and return job data as it comes in

        :returns: all of the information for the JID
        """
        if not isinstance(minions, set):
            if isinstance(minions, str):
                minions = {minions}
            elif isinstance(minions, (list, tuple)):
                minions = set(list(minions))

        if timeout is None:
            timeout = self.opts["timeout"]
        gather_job_timeout = int(
            kwargs.get("gather_job_timeout", self.opts["gather_job_timeout"])
        )
        start = int(time.time())

        # timeouts per minion, id_ -> timeout time
        minion_timeouts = {}

        found = set()
        missing = set()
        # Check to see if the jid is real, if not return the empty dict
        try:
            if not self.returns_for_job(jid):
                log.warning("jid does not exist")
                yield {}
                # stop the iteration, since the jid is invalid
                return
        except Exception as exc:  # pylint: disable=broad-except
            log.warning(
                "Returner unavailable: %s", exc, exc_info_on_loglevel=logging.DEBUG
            )
        # Wait for the hosts to check in
        last_time = False
        # iterator for this job's return
        if self.opts["order_masters"]:
            # If we are a MoM, we need to gather expected minions from downstreams masters.
            ret_iter = self.get_returns_no_block(f"(salt/job|syndic/.*)/{jid}", "regex")
        else:
            ret_iter = self.get_returns_no_block(f"salt/job/{jid}")
        # iterator for the info of this job
        jinfo_iter = []
        # open event jids that need to be un-subscribed from later
        open_jids = set()
        timeout_at = time.time() + timeout
        gather_syndic_wait = time.time() + self.opts["syndic_wait"]
        # are there still minions running the job out there
        # start as True so that we ping at least once
        minions_running = True
        log.debug(
            "get_iter_returns for jid %s sent to %s will timeout at %s",
            jid,
            minions,
            datetime.fromtimestamp(timeout_at).time(),
        )
        while True:
            # Process events until timeout is reached or all minions have returned
            for raw in ret_iter:
                # if we got None, then there were no events
                if raw is None:
                    break
                if "error" in raw.get("data", {}):
                    yield {
                        "error": {
                            "name": "AuthorizationError",
                            "message": "Authorization error occurred.",
                        }
                    }
                if "minions" in raw.get("data", {}):
                    minions.update(raw["data"]["minions"])
                    if "missing" in raw.get("data", {}):
                        missing.update(raw["data"]["missing"])
                    continue

                # Anything below this point is expected to be a job return event.
                if not raw["tag"].startswith(f"salt/job/{jid}/ret/"):
                    log.debug("Skipping non return event: %s", raw["tag"])
                    continue
                if "return" not in raw["data"]:
                    log.warning("Malformed event return: %s", raw["tag"])
                    continue
                if kwargs.get("raw", False):
                    found.add(raw["data"]["id"])
                    yield raw
                else:
                    found.add(raw["data"]["id"])
                    ret = {raw["data"]["id"]: {"ret": raw["data"]["return"]}}
                    if "out" in raw["data"]:
                        ret[raw["data"]["id"]]["out"] = raw["data"]["out"]
                    if "retcode" in raw["data"]:
                        ret[raw["data"]["id"]]["retcode"] = raw["data"]["retcode"]
                    if "jid" in raw["data"]:
                        ret[raw["data"]["id"]]["jid"] = raw["data"]["jid"]
                    if kwargs.get("_cmd_meta", False):
                        ret[raw["data"]["id"]].update(raw["data"])
                    log.debug("jid %s return from %s", jid, raw["data"]["id"])
                    yield ret

            # if we have all of the returns (and we aren't a syndic), no need for anything fancy
            if (
                len(found.intersection(minions)) >= len(minions)
                and not self.opts["order_masters"]
            ):
                # All minions have returned, break out of the loop
                log.debug("jid %s found all minions %s", jid, found)
                break
            elif (
                len(found.intersection(minions)) >= len(minions)
                and self.opts["order_masters"]
            ):
                if (
                    len(found) >= len(minions)
                    and len(minions) > 0
                    and time.time() > gather_syndic_wait
                ):
                    # There were some minions to find and we found them
                    # However, this does not imply that *all* masters have yet responded with expected minion lists.
                    # Therefore, continue to wait up to the syndic_wait period (calculated in gather_syndic_wait) to see
                    # if additional lower-level masters deliver their lists of expected
                    # minions.
                    break
            # If we get here we may not have gathered the minion list yet. Keep waiting
            # for all lower-level masters to respond with their minion lists

            # let start the timeouts for all remaining minions

            for id_ in minions - found:
                # if we have a new minion in the list, make sure it has a timeout
                if id_ not in minion_timeouts:
                    minion_timeouts[id_] = time.time() + timeout

            # if the jinfo has timed out and some minions are still running the job
            # re-do the ping
            if time.time() > timeout_at and minions_running:
                # since this is a new ping, no one has responded yet
                jinfo = self.gather_job_info(
                    jid, list(minions - found), "list", **kwargs
                )
                minions_running = False
                # if we weren't assigned any jid that means the master thinks
                # we have nothing to send
                if "jid" not in jinfo:
                    jinfo_iter = []
                else:
                    jinfo_iter = self.get_returns_no_block(
                        "salt/job/{}".format(jinfo["jid"])
                    )
                timeout_at = time.time() + gather_job_timeout
                # if you are a syndic, wait a little longer
                if self.opts["order_masters"]:
                    timeout_at += self.opts.get("syndic_wait", 1)

            # check for minions that are running the job still
            for raw in jinfo_iter:
                # if there are no more events, lets stop waiting for the jinfo
                if raw is None:
                    break
                try:
                    if raw["data"]["retcode"] > 0:
                        log.error(
                            "saltutil returning errors on minion %s", raw["data"]["id"]
                        )
                        minions.remove(raw["data"]["id"])
                        break
                except KeyError as exc:
                    # This is a safe pass. We're just using the try/except to
                    # avoid having to deep-check for keys.
                    missing_key = str(exc).strip("'\"")
                    if missing_key == "retcode":
                        log.debug("retcode missing from client return")
                    else:
                        log.debug(
                            "Passing on saltutil error. Key '%s' missing "
                            "from client return. This may be an error in "
                            "the client.",
                            missing_key,
                        )
                # Keep track of the jid events to unsubscribe from later
                open_jids.add(jinfo["jid"])

                # TODO: move to a library??
                if "minions" in raw.get("data", {}):
                    minions.update(raw["data"]["minions"])
                    continue
                if "syndic" in raw.get("data", {}):
                    minions.update(raw["syndic"])
                    continue
                if "return" not in raw.get("data", {}):
                    continue

                # if the job isn't running there anymore... don't count
                if raw["data"]["return"] == {}:
                    continue

                # if the minion throws an exception containing the word "return"
                # the master will try to handle the string as a dict in the next
                # step. Check if we have a string, log the issue and continue.
                if isinstance(raw["data"]["return"], str):
                    log.error("unexpected return from minion: %s", raw)
                    continue

                if (
                    "return" in raw["data"]["return"]
                    and raw["data"]["return"]["return"] == {}
                ):
                    continue

                # if we didn't originally target the minion, lets add it to the list
                if raw["data"]["id"] not in minions:
                    minions.add(raw["data"]["id"])
                # update this minion's timeout, as long as the job is still running
                minion_timeouts[raw["data"]["id"]] = time.time() + timeout
                # a minion returned, so we know its running somewhere
                minions_running = True

            # if we have hit gather_job_timeout (after firing the job) AND
            # if we have hit all minion timeouts, lets call it
            now = time.time()
            # if we have finished waiting, and no minions are running the job
            # then we need to see if each minion has timedout
            done = (now > timeout_at) and not minions_running
            if done:
                # if all minions have timeod out
                for id_ in minions - found:
                    if now < minion_timeouts[id_]:
                        done = False
                        break
            if done:
                break

            # don't spin
            if block:
                time.sleep(0.01)
            else:
                yield

        # If there are any remaining open events, clean them up.
        if open_jids:
            for jid in open_jids:
                self.event.unsubscribe(jid)

        if expect_minions:
            for minion in list(minions - found):
                yield {minion: {"failed": True}}

        # Filter out any minions marked as missing for which we received
        # returns (prevents false events sent due to higher-level masters not
        # knowing about lower-level minions).
        missing -= found

        # Report on missing minions
        if missing:
            for minion in missing:
                yield {minion: {"failed": True}}

    def get_returns(self, jid, minions, timeout=None):
        """
        Get the returns for the command line interface via the event system
        """
        minions = set(minions)
        if timeout is None:
            timeout = self.opts["timeout"]
        start = int(time.time())
        timeout_at = start + timeout
        log.debug(
            "get_returns for jid %s sent to %s will timeout at %s",
            jid,
            minions,
            datetime.fromtimestamp(timeout_at).time(),
        )

        found = set()
        ret = {}
        # Check to see if the jid is real, if not return the empty dict
        try:
            if (
                self.returners["{}.get_load".format(self.opts["master_job_cache"])](jid)
                == {}
            ):
                log.warning("jid does not exist")
                return ret
        except Exception as exc:  # pylint: disable=broad-except
            raise SaltClientError(
                "Master job cache returner [{}] failed to verify jid. "
                "Exception details: {}".format(self.opts["master_job_cache"], exc)
            )

        # Wait for the hosts to check in
        while True:
            time_left = timeout_at - int(time.time())
            wait = max(1, time_left)
            raw = self.event.get_event(wait, jid, auto_reconnect=self.auto_reconnect)
            if raw is not None and "return" in raw:
                found.add(raw["id"])
                ret[raw["id"]] = raw["return"]
                if len(found.intersection(minions)) >= len(minions):
                    # All minions have returned, break out of the loop
                    log.debug("jid %s found all minions", jid)
                    break
                continue
            # Then event system timeout was reached and nothing was returned
            if len(found.intersection(minions)) >= len(minions):
                # All minions have returned, break out of the loop
                log.debug("jid %s found all minions", jid)
                break
            if int(time.time()) > timeout_at:
                log.info(
                    "jid %s minions %s did not return in time", jid, (minions - found)
                )
                break
            time.sleep(0.01)
        return ret

    def get_full_returns(self, jid, minions, timeout=None):
        """
        This method starts off a watcher looking at the return data for
        a specified jid, it returns all of the information for the jid
        """
        # TODO: change this from ret to return... or the other way.
        #       Its inconsistent, we should pick one

        ret = {}
        # create the iterator-- since we want to get anyone in the middle
        event_iter = self.get_event_iter_returns(jid, minions, timeout=timeout)

        try:
            data = self.returners["{}.get_jid".format(self.opts["master_job_cache"])](
                jid
            )
        except Exception as exc:  # pylint: disable=broad-except
            raise SaltClientError(
                "Returner {} could not fetch jid data. Exception details: {}".format(
                    self.opts["master_job_cache"], exc
                )
            )
        for minion in data:
            m_data = {}
            if "return" in data[minion]:
                m_data["ret"] = data[minion].get("return")
            else:
                m_data["ret"] = data[minion].get("return")
            if "out" in data[minion]:
                m_data["out"] = data[minion]["out"]
            if minion in ret:
                ret[minion].update(m_data)
            else:
                ret[minion] = m_data

        # if we have all the minion returns, lets just return
        if len(set(ret).intersection(minions)) >= len(minions):
            return ret

        # otherwise lets use the listener we created above to get the rest
        for event_ret in event_iter:
            # if nothing in the event_ret, skip
            if event_ret == {}:
                time.sleep(0.02)
                continue
            for minion, m_data in event_ret.items():
                if minion in ret:
                    ret[minion].update(m_data)
                else:
                    ret[minion] = m_data

            # are we done yet?
            if len(set(ret).intersection(minions)) >= len(minions):
                return ret

        # otherwise we hit the timeout, return what we have
        return ret

    def get_cache_returns(self, jid):
        """
        Execute a single pass to gather the contents of the job cache
        """
        ret = {}

        try:
            data = self.returners["{}.get_jid".format(self.opts["master_job_cache"])](
                jid
            )
        except Exception as exc:  # pylint: disable=broad-except
            raise SaltClientError(
                "Could not examine master job cache. "
                "Error occurred in {} returner. "
                "Exception details: {}".format(self.opts["master_job_cache"], exc)
            )
        for minion in data:
            m_data = {}
            if "return" in data[minion]:
                m_data["ret"] = data[minion].get("return")
            else:
                m_data["ret"] = data[minion].get("return")
            if "out" in data[minion]:
                m_data["out"] = data[minion]["out"]
            if minion in ret:
                ret[minion].update(m_data)
            else:
                ret[minion] = m_data

        return ret

    def get_cli_static_event_returns(
        self,
        jid,
        minions,
        timeout=None,
        tgt="*",
        tgt_type="glob",
        verbose=False,
        show_timeout=False,
        show_jid=False,
    ):
        """
        Get the returns for the command line interface via the event system
        """
        log.trace("entered - function get_cli_static_event_returns()")
        minions = set(minions)
        if verbose:
            msg = f"Executing job with jid {jid}"
            print(msg)
            print("-" * len(msg) + "\n")
        elif show_jid:
            print(f"jid: {jid}")

        if timeout is None:
            timeout = self.opts["timeout"]

        start = int(time.time())
        timeout_at = start + timeout
        found = set()
        ret = {}
        # Check to see if the jid is real, if not return the empty dict
        try:
            if (
                self.returners["{}.get_load".format(self.opts["master_job_cache"])](jid)
                == {}
            ):
                log.warning("jid does not exist")
                return ret
        except Exception as exc:  # pylint: disable=broad-except
            raise SaltClientError(
                "Load could not be retrieved from "
                "returner {}. Exception details: {}".format(
                    self.opts["master_job_cache"], exc
                )
            )
        # Wait for the hosts to check in
        while True:
            # Process events until timeout is reached or all minions have returned
            time_left = timeout_at - int(time.time())
            # Wait 0 == forever, use a minimum of 1s
            wait = max(1, time_left)
            jid_tag = f"salt/job/{jid}"
            raw = self.event.get_event(
                wait, jid_tag, auto_reconnect=self.auto_reconnect
            )
            if raw is not None and "return" in raw:
                if "minions" in raw.get("data", {}):
                    minions.update(raw["data"]["minions"])
                    continue
                found.add(raw["id"])
                ret[raw["id"]] = {"ret": raw["return"]}
                ret[raw["id"]]["success"] = raw.get("success", False)
                if "out" in raw:
                    ret[raw["id"]]["out"] = raw["out"]
                if len(found.intersection(minions)) >= len(minions):
                    # All minions have returned, break out of the loop
                    break
                continue
            # Then event system timeout was reached and nothing was returned
            if len(found.intersection(minions)) >= len(minions):
                # All minions have returned, break out of the loop
                break
            if int(time.time()) > timeout_at:
                if verbose or show_timeout:
                    if self.opts.get("minion_data_cache", False) or tgt_type in (
                        "glob",
                        "pcre",
                        "list",
                    ):
                        if len(found) < len(minions):
                            fail = sorted(list(minions.difference(found)))
                            for minion in fail:
                                ret[minion] = {
                                    "out": "no_return",
                                    "ret": "Minion did not return",
                                }
                break
            time.sleep(0.01)

        self._clean_up_subscriptions(jid)
        return ret

    def get_cli_event_returns(
        self,
        jid,
        minions,
        timeout=None,
        tgt="*",
        tgt_type="glob",
        verbose=False,
        progress=False,
        show_timeout=False,
        show_jid=False,
        **kwargs,
    ):
        """
        Get the returns for the command line interface via the event system
        """
        log.trace("func get_cli_event_returns()")

        if verbose:
            msg = f"Executing job with jid {jid}"
            print(msg)
            print("-" * len(msg) + "\n")
        elif show_jid:
            print(f"jid: {jid}")

        # lazy load the connected minions
        connected_minions = None
        return_count = 0

        for ret in self.get_iter_returns(
            jid,
            minions,
            timeout=timeout,
            tgt=tgt,
            tgt_type=tgt_type,
            # (gtmanfred) expect_minions is popped here in case it is passed from a client
            # call. If this is not popped, then it would be passed twice to
            # get_iter_returns.
            expect_minions=(
                kwargs.pop("expect_minions", False) or verbose or show_timeout
            ),
            **kwargs,
        ):
            log.debug("return event: %s", ret)
            return_count = return_count + 1
            if progress:
                for id_, min_ret in ret.items():
                    if not min_ret.get("failed") is True:
                        yield {
                            "minion_count": len(minions),
                            "return_count": return_count,
                        }
            # replace the return structure for missing minions
            for id_, min_ret in ret.items():
                if min_ret.get("failed") is True:
                    if connected_minions is None:
                        connected_minions = salt.utils.minions.CkMinions(
                            self.opts
                        ).connected_ids()
                    if (
                        self.opts["minion_data_cache"]
                        and salt.cache.factory(self.opts).contains(
                            f"minions/{id_}", "data"
                        )
                        and connected_minions
                        and id_ not in connected_minions
                    ):

                        yield {
                            id_: {
                                "out": "no_return",
                                "ret": "Minion did not return. [Not connected]",
                                "retcode": salt.defaults.exitcodes.EX_GENERIC,
                            }
                        }
                    else:
                        # don't report syndics as unresponsive minions
                        if not os.path.exists(
                            os.path.join(self.opts["syndic_dir"], id_)
                        ):
                            yield {
                                id_: {
                                    "out": "no_return",
                                    "ret": (
                                        "Minion did not return. [No response]\nThe"
                                        " minions may not have all finished running and"
                                        " any remaining minions will return upon"
                                        " completion. To look up the return data for"
                                        " this job later, run the following"
                                        " command:\n\nsalt-run jobs.lookup_jid {}".format(
                                            jid
                                        )
                                    ),
                                    "retcode": salt.defaults.exitcodes.EX_GENERIC,
                                }
                            }
                elif "error" in min_ret:
                    raise AuthorizationError("Authorization error occurred")
                else:
                    yield {id_: min_ret}

        self._clean_up_subscriptions(jid)

    def get_event_iter_returns(self, jid, minions, timeout=None):
        """
        Gather the return data from the event system, break hard when timeout
        is reached.
        """
        log.trace("entered - function get_event_iter_returns()")
        if timeout is None:
            timeout = self.opts["timeout"]

        timeout_at = time.time() + timeout

        found = set()
        # Check to see if the jid is real, if not return the empty dict
        if (
            self.returners["{}.get_load".format(self.opts["master_job_cache"])](jid)
            == {}
        ):
            log.warning("jid does not exist")
            yield {}
            # stop the iteration, since the jid is invalid
            return
        # Wait for the hosts to check in
        while True:
            raw = self.event.get_event(timeout, auto_reconnect=self.auto_reconnect)
            if raw is None or time.time() > timeout_at:
                # Timeout reached
                break
            if "minions" in raw.get("data", {}):
                continue
            try:
                # There might be two jobs for the same minion, so we have to check for the jid
                if jid == raw["jid"]:
                    found.add(raw["id"])
                    ret = {raw["id"]: {"ret": raw["return"]}}
                else:
                    continue
            except KeyError:
                # Ignore other erroneous messages
                continue
            if "out" in raw:
                ret[raw["id"]]["out"] = raw["out"]
            yield ret
            time.sleep(0.02)

    def _resolve_nodegroup(self, ng):
        """
        Resolve a nodegroup into its configured components
        """
        if ng not in self.opts["nodegroups"]:
            conf_file = self.opts.get("conf_file", "the master config file")
            raise SaltInvocationError(f"Node group {ng} unavailable in {conf_file}")
        return salt.utils.minions.nodegroup_comp(ng, self.opts["nodegroups"])

    def _prep_pub(self, tgt, fun, arg, tgt_type, ret, jid, timeout, **kwargs):
        """
        Set up the payload_kwargs to be sent down to the master
        """
        if tgt_type == "nodegroup":
            tgt = self._resolve_nodegroup(tgt)
            tgt_type = "compound"

        if tgt_type == "compound":
            #  Resolve all nodegroups, so that the minions don't have to.
            new_tgt = list()
            log.debug("compound resolution: original tgt: %s", tgt)

            if isinstance(tgt, str):
                tgt = tgt.split()

            for word in tgt:
                if word.startswith("N@") and len(word) > 2:
                    resolved = self._resolve_nodegroup(word[2:])
                    new_tgt.extend(resolved)
                else:
                    new_tgt.append(word)

            log.debug("compound resolution: new_tgt: %s", new_tgt)
            tgt = " ".join(new_tgt)

        # Convert a range expression to a list of nodes and change expression
        # form to list
        if tgt_type == "range" and HAS_RANGE:
            tgt = self._convert_range_to_list(tgt)
            tgt_type = "list"

        # If an external job cache is specified add it to the ret list
        if self.opts.get("ext_job_cache"):
            if ret:
                ret += ",{}".format(self.opts["ext_job_cache"])
            else:
                ret = self.opts["ext_job_cache"]

        # format the payload - make a function that does this in the payload
        #   module

        # Generate the standard keyword args to feed to format_payload
        payload_kwargs = {
            "cmd": "publish",
            "tgt": tgt,
            "fun": fun,
            "arg": arg,
            "key": self.key,
            "tgt_type": tgt_type,
            "ret": ret,
            "jid": jid,
        }

        # if kwargs are passed, pack them.
        if kwargs:
            payload_kwargs["kwargs"] = kwargs

        payload_kwargs["user"] = self.salt_user

        # If we're a syndication master, pass the timeout
        if self.opts["order_masters"]:
            payload_kwargs["to"] = timeout

        return payload_kwargs

    def pub(
        self,
        tgt,
        fun,
        arg=(),
        tgt_type="glob",
        ret="",
        jid="",
        timeout=5,
        listen=False,
        **kwargs,
    ):
        """
        Take the required arguments and publish the given command.
        Arguments:
            tgt:
                The tgt is a regex or a glob used to match up the ids on
                the minions. Salt works by always publishing every command
                to all of the minions and then the minions determine if
                the command is for them based on the tgt value.
            fun:
                The function name to be called on the remote host(s), this
                must be a string in the format "<modulename>.<function name>"
            arg:
                The arg option needs to be a tuple of arguments to pass
                to the calling function, if left blank
        Returns:
            jid:
                A string, as returned by the publisher, which is the job
                id, this will inform the client where to get the job results
            minions:
                A set, the targets that the tgt passed should match.
        """
        # Make sure the publisher is running by checking the unix socket
        if self.opts.get("ipc_mode", "") != "tcp" and not os.path.exists(
            os.path.join(self.opts["sock_dir"], "publish_pull.ipc")
        ):
            log.error(
                "Unable to connect to the salt master publisher at %s",
                self.opts["sock_dir"],
            )
            raise SaltClientError

        payload_kwargs = self._prep_pub(
            tgt, fun, arg, tgt_type, ret, jid, timeout, **kwargs
        )

        master_uri = "tcp://{}:{}".format(
            salt.utils.network.ip_bracket(self.opts["interface"]),
            str(self.opts["ret_port"]),
        )

        with salt.channel.client.ReqChannel.factory(
            self.opts, crypt="clear", master_uri=master_uri
        ) as channel:
            try:
                # Ensure that the event subscriber is connected.
                # If not, we won't get a response, so error out
                if listen and not self.event.connect_pub(timeout=timeout):
                    raise SaltReqTimeoutError()
                payload = channel.send(payload_kwargs, timeout=timeout)
            except SaltReqTimeoutError as err:
                log.error(err)
                raise SaltReqTimeoutError(
                    "Salt request timed out. The master is not responding. You "
                    "may need to run your command with `--async` in order to "
                    "bypass the congested event bus. With `--async`, the CLI tool "
                    "will print the job id (jid) and exit immediately without "
                    "listening for responses. You can then use "
                    "`salt-run jobs.lookup_jid` to look up the results of the job "
                    "in the job cache later."
                )

            if not payload:
                # The master key could have changed out from under us! Regen
                # and try again if the key has changed
                key = self.__read_master_key()
                if key == self.key:
                    return payload
                self.key = key
                payload_kwargs["key"] = self.key
                payload = channel.send(payload_kwargs)

            error = payload.pop("error", None)
            if error is not None:
                if isinstance(error, dict):
                    err_name = error.get("name", "")
                    err_msg = error.get("message", "")
                    if err_name == "AuthenticationError":
                        raise AuthenticationError(err_msg)
                    elif err_name == "AuthorizationError":
                        raise AuthorizationError(err_msg)

                raise PublishError(error)

            if not payload:
                return payload

        return {"jid": payload["load"]["jid"], "minions": payload["load"]["minions"]}

    @salt.ext.tornado.gen.coroutine
    def pub_async(
        self,
        tgt,
        fun,
        arg=(),
        tgt_type="glob",
        ret="",
        jid="",
        timeout=5,
        io_loop=None,
        listen=True,
        **kwargs,
    ):
        """
        Take the required arguments and publish the given command.
        Arguments:
            tgt:
                The tgt is a regex or a glob used to match up the ids on
                the minions. Salt works by always publishing every command
                to all of the minions and then the minions determine if
                the command is for them based on the tgt value.
            fun:
                The function name to be called on the remote host(s), this
                must be a string in the format "<modulename>.<function name>"
            arg:
                The arg option needs to be a tuple of arguments to pass
                to the calling function, if left blank
        Returns:
            jid:
                A string, as returned by the publisher, which is the job
                id, this will inform the client where to get the job results
            minions:
                A set, the targets that the tgt passed should match.
        """
        # Make sure the publisher is running by checking the unix socket
        if self.opts.get("ipc_mode", "") != "tcp" and not os.path.exists(
            os.path.join(self.opts["sock_dir"], "publish_pull.ipc")
        ):
            log.error(
                "Unable to connect to the salt master publisher at %s",
                self.opts["sock_dir"],
            )
            raise SaltClientError

        payload_kwargs = self._prep_pub(
            tgt, fun, arg, tgt_type, ret, jid, timeout, **kwargs
        )

        master_uri = (
            "tcp://"
            + salt.utils.network.ip_bracket(self.opts["interface"])
            + ":"
            + str(self.opts["ret_port"])
        )

        with salt.channel.client.AsyncReqChannel.factory(
            self.opts, io_loop=io_loop, crypt="clear", master_uri=master_uri
        ) as channel:
            try:
                # Ensure that the event subscriber is connected.
                # If not, we won't get a response, so error out
                if listen and not self.event.connect_pub(timeout=timeout):
                    raise SaltReqTimeoutError()
                payload = yield channel.send(payload_kwargs, timeout=timeout)
            except SaltReqTimeoutError:
                raise SaltReqTimeoutError(
                    "Salt request timed out. The master is not responding. You "
                    "may need to run your command with `--async` in order to "
                    "bypass the congested event bus. With `--async`, the CLI tool "
                    "will print the job id (jid) and exit immediately without "
                    "listening for responses. You can then use "
                    "`salt-run jobs.lookup_jid` to look up the results of the job "
                    "in the job cache later."
                )

            if not payload:
                # The master key could have changed out from under us! Regen
                # and try again if the key has changed
                key = self.__read_master_key()
                if key == self.key:
                    raise salt.ext.tornado.gen.Return(payload)
                self.key = key
                payload_kwargs["key"] = self.key
                payload = yield channel.send(payload_kwargs)

            error = payload.pop("error", None)
            if error is not None:
                if isinstance(error, dict):
                    err_name = error.get("name", "")
                    err_msg = error.get("message", "")
                    if err_name == "AuthenticationError":
                        raise AuthenticationError(err_msg)
                    elif err_name == "AuthorizationError":
                        raise AuthorizationError(err_msg)

                raise PublishError(error)

            if not payload:
                raise salt.ext.tornado.gen.Return(payload)

        raise salt.ext.tornado.gen.Return(
            {"jid": payload["load"]["jid"], "minions": payload["load"]["minions"]}
        )

    # pylint: disable=W1701
    def __del__(self):
        # This IS really necessary!
        # When running tests, if self.events is not destroyed, we leak 2
        # threads per test case which uses self.client
        self.destroy()

    # pylint: enable=W1701

    def _clean_up_subscriptions(self, job_id):
        if self.opts.get("order_masters"):
            self.event.unsubscribe(f"syndic/.*/{job_id}", "regex")
        self.event.unsubscribe(f"salt/job/{job_id}")

    def destroy(self):
        if self.event is not None:
            self.event.destroy()
            self.event = None

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.destroy()


class FunctionWrapper(dict):
    """
    Create a function wrapper that looks like the functions dict on the minion
    but invoked commands on the minion via a LocalClient.

    This allows SLS files to be loaded with an object that calls down to the
    minion when the salt functions dict is referenced.
    """

    def __init__(self, opts, minion):
        super().__init__()
        self.opts = opts
        self.minion = minion
        self.local = LocalClient(self.opts["conf_file"])
        self.functions = self.__load_functions()

    def __missing__(self, key):
        """
        Since the function key is missing, wrap this call to a command to the
        minion of said key if it is available in the self.functions set
        """
        if key not in self.functions:
            raise KeyError
        return self.run_key(key)

    def __load_functions(self):
        """
        Find out what functions are available on the minion
        """
        return set(
            self.local.cmd(self.minion, "sys.list_functions").get(self.minion, [])
        )

    def run_key(self, key):
        """
        Return a function that executes the arguments passed via the local
        client
        """

        def func(*args, **kwargs):
            """
            Run a remote call
            """
            args = list(args)
            for _key, _val in kwargs.items():
                args.append(f"{_key}={_val}")
            return self.local.cmd(self.minion, key, args)

        return func


class Caller:
    """
    ``Caller`` is the same interface used by the :command:`salt-call`
    command-line tool on the Salt Minion.

    .. versionchanged:: 2015.8.0
        Added the ``cmd`` method for consistency with the other Salt clients.
        The existing ``function`` and ``sminion.functions`` interfaces still
        exist but have been removed from the docs.

    Importing and using ``Caller`` must be done on the same machine as a
    Salt Minion and it must be done using the same user that the Salt Minion is
    running as.

    Usage:

    .. code-block:: python

        import salt.client
        caller = salt.client.Caller()
        caller.cmd('test.ping')

    Note, a running master or minion daemon is not required to use this class.
    Running ``salt-call --local`` simply sets :conf_minion:`file_client` to
    ``'local'``. The same can be achieved at the Python level by including that
    setting in a minion config file.

    .. versionadded:: 2014.7.0
        Pass the minion config as the ``mopts`` dictionary.

    .. code-block:: python

        import salt.client
        import salt.config
        __opts__ = salt.config.minion_config('/etc/salt/minion')
        __opts__['file_client'] = 'local'
        caller = salt.client.Caller(mopts=__opts__)
    """

    def __init__(self, c_path=os.path.join(syspaths.CONFIG_DIR, "minion"), mopts=None):
        # Late-import of the minion module to keep the CLI as light as possible
        import salt.minion

        if mopts:
            self.opts = mopts
        else:
            self.opts = salt.config.minion_config(c_path)
        self.sminion = salt.minion.SMinion(self.opts)

    def cmd(self, fun, *args, **kwargs):
        """
        Call an execution module with the given arguments and keyword arguments

        .. versionchanged:: 2015.8.0
            Added the ``cmd`` method for consistency with the other Salt clients.
            The existing ``function`` and ``sminion.functions`` interfaces still
            exist but have been removed from the docs.

        .. code-block:: python

            caller.cmd('test.arg', 'Foo', 'Bar', baz='Baz')

            caller.cmd('event.send', 'myco/myevent/something',
                data={'foo': 'Foo'}, with_env=['GIT_COMMIT'], with_grains=True)
        """
        return self.sminion.functions[fun](*args, **kwargs)

    def function(self, fun, *args, **kwargs):
        """
        Call a single salt function
        """
        func = self.sminion.functions[fun]
        args, kwargs = salt.minion.load_args_and_kwargs(
            func, salt.utils.args.parse_input(args), kwargs
        )
        return func(*args, **kwargs)


class ProxyCaller:
    """
    ``ProxyCaller`` is the same interface used by the :command:`salt-call`
    with the args ``--proxyid <proxyid>`` command-line tool on the Salt Proxy
    Minion.

    Importing and using ``ProxyCaller`` must be done on the same machine as a
    Salt Minion and it must be done using the same user that the Salt Minion is
    running as.

    Usage:

    .. code-block:: python

        import salt.client
        caller = salt.client.ProxyCaller()
        caller.cmd('test.ping')

    Note, a running master or minion daemon is not required to use this class.
    Running ``salt-call --local`` simply sets :conf_minion:`file_client` to
    ``'local'``. The same can be achieved at the Python level by including that
    setting in a minion config file.

    .. code-block:: python

        import salt.client
        import salt.config
        __opts__ = salt.config.proxy_config('/etc/salt/proxy', minion_id='quirky_edison')
        __opts__['file_client'] = 'local'
        caller = salt.client.ProxyCaller(mopts=__opts__)

    .. note::

        To use this for calling proxies, the :py:func:`is_proxy functions
        <salt.utils.platform.is_proxy>` requires that ``--proxyid`` be an
        argument on the commandline for the script this is used in, or that the
        string ``proxy`` is in the name of the script.
    """

    def __init__(self, c_path=os.path.join(syspaths.CONFIG_DIR, "proxy"), mopts=None):
        # Late-import of the minion module to keep the CLI as light as possible
        import salt.minion

        self.opts = mopts or salt.config.proxy_config(c_path)
        self.sminion = salt.minion.SProxyMinion(self.opts)

    def cmd(self, fun, *args, **kwargs):
        """
        Call an execution module with the given arguments and keyword arguments

        .. code-block:: python

            caller.cmd('test.arg', 'Foo', 'Bar', baz='Baz')

            caller.cmd('event.send', 'myco/myevent/something',
                data={'foo': 'Foo'}, with_env=['GIT_COMMIT'], with_grains=True)
        """
        func = self.sminion.functions[fun]
        data = {"arg": args, "fun": fun}
        data.update(kwargs)
        executors = getattr(self.sminion, "module_executors", []) or self.opts.get(
            "module_executors", ["direct_call"]
        )
        if isinstance(executors, str):
            executors = [executors]
        for name in executors:
            fname = f"{name}.execute"
            if fname not in self.sminion.executors:
                raise SaltInvocationError(f"Executor '{name}' is not available")
            return_data = self.sminion.executors[fname](
                self.opts, data, func, args, kwargs
            )
            if return_data is not None:
                break
        return return_data



"""
    :codeauthor: Rupesh Tare <rupesht@saltstack.com>
"""

import os.path

import pytest

import salt.modules.devmap as devmap
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {devmap: {}}


def test_multipath_list():
    """
    Test for Device-Mapper Multipath list
    """
    mock = MagicMock(return_value="A")
    with patch.dict(devmap.__salt__, {"cmd.run": mock}):
        assert devmap.multipath_list() == ["A"]


def test_multipath_flush():
    """
    Test for Device-Mapper Multipath flush
    """
    mock = MagicMock(return_value=False)
    with patch.object(os.path, "exists", mock):
        assert devmap.multipath_flush("device") == "device does not exist"

    mock = MagicMock(return_value=True)
    with patch.object(os.path, "exists", mock):
        mock = MagicMock(return_value="A")
        with patch.dict(devmap.__salt__, {"cmd.run": mock}):
            assert devmap.multipath_flush("device") == ["A"]
